class
	P_SRC_ENV_STATIC

inherit
	CE_UTILITIES
	S_SYS_SOCKET_STATIC
	S_BITS_ERRNO_STATIC
	S_STRING_STATIC
	S_STDIO_STATIC
	S_NETDB_STATIC
	S_UNISTD_STATIC

feature {ANY} -- public routines

	env_sock_init: INTEGER_32
		do
			Result := 0
		end

	env_write_socket (a_conn: CE_POINTER [PD_MONGO]; a_buf: CE_POINTER [ANY]; a_len: INTEGER_32): INTEGER_32
		local
			l_len: INTEGER_32
			l_cbuf: CE_POINTER [INTEGER_8]
			l_flags: INTEGER_32
			l_sent: INTEGER_32
			l_tmp_1: CE_POINTER [INTEGER_32]
			l_return: BOOLEAN
		do
			l_len := a_len
			l_cbuf := create {CE_INTEGER_8_POINTER}.make_cast (a_buf)
			l_flags := 16_384
			from
			until
				not (not (l_return) and then (l_len.to_boolean))
			loop
				l_sent := (send (a_conn.item.sock, create {CE_VOID_POINTER}.make_cast (l_cbuf), l_len.to_natural_32, l_flags))
				if l_sent = -1 then
					if (errno_location).item = 32 then
						a_conn.item.connected := 0
					end
					l_tmp_1 := errno_location
					a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_io_error, (strerror ((errno_location).item)), l_tmp_1.item)
					Result := (-1)
					l_return := True
				end
				if not l_return then
					l_cbuf := l_cbuf + l_sent
					l_len := l_len - l_sent
				end
			end
			if not l_return then
				Result := 0
			end
		end

	env_set_socket_op_timeout (a_conn: CE_POINTER [PD_MONGO]; a_millis: INTEGER_32): INTEGER_32
		local
			l_tv: CE_ARRAY [SD_TIMEVAL]
			l_return: BOOLEAN
		do
			create {CE_CLASS_ARRAY [SD_TIMEVAL]} l_tv.make (<<1>>)

			l_tv.item.tv_sec := (a_millis // 1_000).to_integer_32
			l_tv.item.tv_usec := ((a_millis \\ 1_000) * 1_000).to_integer_32
			if (setsockopt (a_conn.item.sock, 1, 20, create {CE_VOID_POINTER}.make_cast (l_tv), (create {SD_TIMEVAL}.make).structure_size.to_natural_32)) = -1 then
				a_conn.item.err := {PC_MONGO_ERROR_T}.Mongo_io_error;
				a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_io_error, ce_cstring ("setsockopt SO_RCVTIMEO failed."), (errno_location).item)
				Result := (-1)
				l_return := True
			end
			if not l_return then
				if (setsockopt (a_conn.item.sock, 1, 21, create {CE_VOID_POINTER}.make_cast (l_tv), (create {SD_TIMEVAL}.make).structure_size.to_natural_32)) = -1 then
					a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_io_error, ce_cstring ("setsockopt SO_SNDTIMEO failed."), (errno_location).item)
					Result := (-1)
					l_return := True
				end
			end
			if not l_return then
				Result := 0
			end
		end

	env_socket_connect (a_conn: CE_POINTER [PD_MONGO]; a_host: CE_POINTER [INTEGER_8]; a_port: INTEGER_32): INTEGER_32
		local
			l_port_str: CE_ARRAY [INTEGER_8]
			l_status: INTEGER_32
			l_ai_hints: CE_ARRAY [SD_ADDRINFO]
			l_ai_list: CE_ARRAY [CE_POINTER [SD_ADDRINFO]]
			l_ai_ptr: CE_POINTER [SD_ADDRINFO]
			l_flag: CE_ARRAY [INTEGER_32]
			l_return, l_break, l_continue: BOOLEAN
		do
			create {CE_INTEGER_8_ARRAY} l_port_str.make (<<32>>)
			create {CE_CLASS_ARRAY [SD_ADDRINFO]} l_ai_hints.make (<<1>>)
			create {CE_POINTER_ARRAY [SD_ADDRINFO]} l_ai_list.make (<<1>>)
			create {CE_INTEGER_32_ARRAY} l_flag.make (<<1>>)

			l_ai_list.item := create {CE_CLASS_POINTER [SD_ADDRINFO]}.make_shared (default_pointer)
			l_ai_ptr := create {CE_CLASS_POINTER [SD_ADDRINFO]}.make_shared (default_pointer)
			a_conn.item.sock := 0
			a_conn.item.connected := 0
			sprintf ([l_port_str, ce_cstring ("%%d"), a_port]).do_nothing;
			(create {P_SRC_BSON_STATIC}).bson_sprintf.item.call ([l_port_str, ce_cstring ("%%d"), a_port])
			memset (create {CE_VOID_POINTER}.make_cast (l_ai_hints), 0, (create {SD_ADDRINFO}.make).structure_size.to_natural_32).do_nothing
			l_ai_hints.item.ai_flags := 32
			l_ai_hints.item.ai_family := 0
			l_ai_hints.item.ai_socktype := 1
			l_status := getaddrinfo (a_host, l_port_str, l_ai_hints, l_ai_list)
			if l_status /= 0 then
				(create {P_SRC_BSON_STATIC}).bson_errprintf.item.call ([ce_cstring ("getaddrinfo failed: %%s"), (gai_strerror (l_status))])
				a_conn.item.err := {PC_MONGO_ERROR_T}.Mongo_conn_addr_fail
				Result := (-1)
				l_return := True
			end
			if not l_return then
				l_ai_ptr := l_ai_list.item
				from
				until
					not (not (l_break) and then (not l_ai_ptr.is_void))
				loop
					l_break := False
					l_continue := False
					a_conn.item.sock := socket (l_ai_ptr.item.ai_family, l_ai_ptr.item.ai_socktype, l_ai_ptr.item.ai_protocol)
					if a_conn.item.sock < 0 then
						a_conn.item.sock := 0
						l_continue := True
					end
					if not (l_break or l_continue) then
						l_status := connect (a_conn.item.sock, l_ai_ptr.item.ai_addr, l_ai_ptr.item.ai_addrlen)
						if l_status /= 0 then
							env_close_socket (a_conn.item.sock).do_nothing
							a_conn.item.sock := 0
							l_continue := True
						end
					end
					if not (l_break or l_continue) then
						if l_ai_ptr.item.ai_protocol = 6 then
							l_flag.item := 1
							setsockopt (a_conn.item.sock, 6, 1, create {CE_VOID_POINTER}.make_cast (l_flag), {PLATFORM}.Integer_32_bytes.to_natural_32).do_nothing
							if a_conn.item.op_timeout_ms > 0 then
								env_set_socket_op_timeout (a_conn, a_conn.item.op_timeout_ms).do_nothing
							end
						end
						a_conn.item.connected := 1
						l_break := True
					end
					if (not l_break) then
						l_ai_ptr := l_ai_ptr.item.ai_next
					end
				end
				l_break := False
				l_continue := False
				freeaddrinfo (l_ai_list.item)
				if not a_conn.item.connected.to_boolean then
					a_conn.item.err := {PC_MONGO_ERROR_T}.Mongo_conn_fail
					Result := (-1)
					l_return := True
				end
			end
			if not l_return then
				Result := 0
			end
		end

	env_close_socket (a_socket_0: INTEGER_32): INTEGER_32
		do
			Result := (close (a_socket_0))
		end

	env_read_socket (a_conn: CE_POINTER [PD_MONGO]; a_buf: CE_POINTER [ANY]; a_len: INTEGER_32): INTEGER_32
		local
			l_len: INTEGER_32
			l_cbuf: CE_POINTER [INTEGER_8]
			l_sent: INTEGER_32
			l_tmp_0: CE_POINTER [INTEGER_32]
			l_tmp_1: CE_POINTER [INTEGER_32]
			l_tmp_2: CE_POINTER [INTEGER_8]
			l_return: BOOLEAN
		do
			l_len := a_len
			l_cbuf := create {CE_INTEGER_8_POINTER}.make_cast (a_buf)
			from
			until
				not (not (l_return) and then (l_len.to_boolean))
			loop
				l_sent := (recv (a_conn.item.sock, create {CE_VOID_POINTER}.make_cast (l_cbuf), l_len.to_natural_32, 0))
				if l_sent = 0 then
					l_tmp_0 := errno_location
					l_tmp_1 := errno_location
					l_tmp_2 := strerror (l_tmp_1.item)
					a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_io_error, l_tmp_2, l_tmp_0.item)
					Result := (-1)
					l_return := True
				elseif l_sent = -1 then
					l_tmp_0 := errno_location
					l_tmp_1 := errno_location
					l_tmp_2 := strerror (l_tmp_1.item)
					a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_io_error, l_tmp_2, l_tmp_0.item)
					Result := (-1)
					l_return := True
				end
				if not l_return then
					l_cbuf := l_cbuf + l_sent
					l_len := l_len - l_sent
				end
			end
			if not l_return then
				Result := 0
			end
		end

end
