class
	SD_IOVEC

inherit
	CE_UTILITIES
		redefine
			copy,
			is_equal
		end
	CE_CLASS
		redefine
			make,
			make_shared,
			structure_size,
			offset_of,
			copy,
			is_equal
		end

create
	make,
	make_shared,
	make_initialized

feature {NONE}

	make
		do
			internal_c_pointer := internal_c_pointer.memory_alloc (structure_size)
			is_allocated_memory := True
		end

	make_shared (a_c_pointer: POINTER)
		do
			internal_c_pointer := a_c_pointer
		end

	make_initialized (a_data: TUPLE [
				a_iov_base: CE_POINTER [ANY]
				a_iov_len: NATURAL_32
			])
		require else
			a_data /= Void
		do
			make
			Current.iov_base := a_data.a_iov_base
			Current.iov_len := a_data.a_iov_len
		end

feature {ANY} -- fields

	iov_base: CE_POINTER [ANY] assign set__iov_base
		do
			create {CE_VOID_POINTER} Result.make_shared (c__iov_base (internal_c_pointer))
		end

	iov_len: NATURAL_32 assign set__iov_len
		do
			Result := c__iov_len (internal_c_pointer)
		end

feature {ANY} -- field setters

	set__iov_base (a_iov_base: CE_POINTER [ANY])
		do
			c__set_iov_base (internal_c_pointer, a_iov_base.c_pointer)
		end

	set__iov_len (a_iov_len: NATURAL_32)
		do
			c__set_iov_len (internal_c_pointer, a_iov_len)
		end

feature {ANY}

	structure_size: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return sizeof(struct iovec);"
		end

	structure_alignment: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return __alignof__(struct iovec);"
		end

	offset_of (a_field_name: STRING_8): INTEGER_32
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_field_name)
			Result := c__offset_of (l_c_str.item)
		end

feature {ANY}

	copy (a_other: like Current)
		do
			make
			internal_c_pointer.memory_copy (a_other.c_pointer, structure_size)
		end

	is_equal (a_other: like Current): BOOLEAN
		do
			Result := internal_c_pointer.memory_compare (a_other.c_pointer, structure_size)
		end

feature {NONE}

	structure_info: CE_STRUCTURE
		do
			-- dummy for CE_CLASS invariant
			create Result.make_struct (<<>>)
			Result.alignment := structure_alignment
		end

	c__iov_base (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct iovec *)$a_c_pointer)->iov_base;"
		end

	c__iov_len (a_c_pointer: POINTER): NATURAL_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct iovec *)$a_c_pointer)->iov_len;"
		end

	c__set_iov_base (a_c_pointer: POINTER; a_iov_base: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct iovec *)$a_c_pointer)->iov_base = $a_iov_base;"
		end

	c__set_iov_len (a_c_pointer: POINTER; a_iov_len: NATURAL_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct iovec *)$a_c_pointer)->iov_len = $a_iov_len;"
		end

	c__offset_of (a_field_name: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%", <string.h>, <stddef.h>, <stdio.h>"
		alias
			"[
				if(strcmp($a_field_name, "iov_base") == 0) return offsetof(struct iovec, iov_base);
				if(strcmp($a_field_name, "iov_len") == 0) return offsetof(struct iovec, iov_len);
				fprintf(stderr, "struct iovec.%s not found\n", (char *)$a_field_name);
				return -1;
			]"
		end

end
