class
	SD_TIMEVAL

inherit
	CE_UTILITIES
		redefine
			copy,
			is_equal
		end
	CE_CLASS
		redefine
			make,
			make_shared,
			structure_size,
			offset_of,
			copy,
			is_equal
		end

create
	make,
	make_shared,
	make_initialized

feature {NONE}

	make
		do
			internal_c_pointer := internal_c_pointer.memory_alloc (structure_size)
			is_allocated_memory := True
		end

	make_shared (a_c_pointer: POINTER)
		do
			internal_c_pointer := a_c_pointer
		end

	make_initialized (a_data: TUPLE [
				a_tv_sec: INTEGER_32
				a_tv_usec: INTEGER_32
			])
		require else
			a_data /= Void
		do
			make
			Current.tv_sec := a_data.a_tv_sec
			Current.tv_usec := a_data.a_tv_usec
		end

feature {ANY} -- fields

	tv_sec: INTEGER_32 assign set__tv_sec
		do
			Result := c__tv_sec (internal_c_pointer)
		end

	tv_usec: INTEGER_32 assign set__tv_usec
		do
			Result := c__tv_usec (internal_c_pointer)
		end

feature {ANY} -- field setters

	set__tv_sec (a_tv_sec: INTEGER_32)
		do
			c__set_tv_sec (internal_c_pointer, a_tv_sec)
		end

	set__tv_usec (a_tv_usec: INTEGER_32)
		do
			c__set_tv_usec (internal_c_pointer, a_tv_usec)
		end

feature {ANY}

	structure_size: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return sizeof(struct timeval);"
		end

	structure_alignment: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return __alignof__(struct timeval);"
		end

	offset_of (a_field_name: STRING_8): INTEGER_32
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_field_name)
			Result := c__offset_of (l_c_str.item)
		end

feature {ANY}

	copy (a_other: like Current)
		do
			make
			internal_c_pointer.memory_copy (a_other.c_pointer, structure_size)
		end

	is_equal (a_other: like Current): BOOLEAN
		do
			Result := internal_c_pointer.memory_compare (a_other.c_pointer, structure_size)
		end

feature {NONE}

	structure_info: CE_STRUCTURE
		do
			-- dummy for CE_CLASS invariant
			create Result.make_struct (<<>>)
			Result.alignment := structure_alignment
		end

	c__tv_sec (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct timeval *)$a_c_pointer)->tv_sec;"
		end

	c__tv_usec (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct timeval *)$a_c_pointer)->tv_usec;"
		end

	c__set_tv_sec (a_c_pointer: POINTER; a_tv_sec: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct timeval *)$a_c_pointer)->tv_sec = $a_tv_sec;"
		end

	c__set_tv_usec (a_c_pointer: POINTER; a_tv_usec: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct timeval *)$a_c_pointer)->tv_usec = $a_tv_usec;"
		end

	c__offset_of (a_field_name: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%", <string.h>, <stddef.h>, <stdio.h>"
		alias
			"[
				if(strcmp($a_field_name, "tv_sec") == 0) return offsetof(struct timeval, tv_sec);
				if(strcmp($a_field_name, "tv_usec") == 0) return offsetof(struct timeval, tv_usec);
				fprintf(stderr, "struct timeval.%s not found\n", (char *)$a_field_name);
				return -1;
			]"
		end

end
