class
	P_SRC_MONGO_STATIC

inherit
	CE_UTILITIES
	S_STDLIB_STATIC
	S_STRING_STATIC

feature {ANY} -- public attributes

	zero: CE_STATIC [CE_ARRAY [INTEGER_32]]
		once ("PROCESS")
			create Result.make (create {CE_INTEGER_32_ARRAY}.make_initialized (<<
					0
				>>))
		end

	one: CE_STATIC [CE_ARRAY [INTEGER_32]]
		once ("PROCESS")
			create Result.make (create {CE_INTEGER_32_ARRAY}.make_initialized (<<
					1
				>>))
		end

	hex: CE_STATIC [CE_ARRAY [INTEGER_8]]
		once ("PROCESS")
			create Result.make (create {CE_INTEGER_8_ARRAY}.make (<<16>>))
			across "0123456789abcdef" as c loop
				Result.item[c.target_index - 1] := c.item.code.to_integer_8
			end
		end

feature {ANY} -- public routines

	mongo_create: CE_POINTER [PD_MONGO]
		do
			Result := (create {CE_CLASS_POINTER [PD_MONGO]}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (((create {PD_MONGO}.make).structure_size.to_natural_32).to_integer_32))))
		end

	mongo_dispose (a_conn: CE_POINTER [PD_MONGO])
		do
			free (create {CE_VOID_POINTER}.make_cast (a_conn))
		end

	mongo_get_primary (a_conn: CE_POINTER [PD_MONGO]): CE_POINTER [INTEGER_8]
		require
			not a_conn.is_void
		local
			l_conn: CE_POINTER [PD_MONGO]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_8_POINTER} Result.make

			l_conn := a_conn
			if not (l_conn.item.connected).to_boolean then
				Result := (create {CE_INTEGER_8_POINTER}.make_shared (default_pointer))
				l_return := True
			elseif ((l_conn.item.primary).item.host).is_void then
				Result := (create {CE_INTEGER_8_POINTER}.make_shared (default_pointer))
				l_return := True
			end
			if not l_return then
				Result := ((l_conn.item.primary).item.get_host_port)
			end
		end

	mongo_get_socket (a_conn: CE_POINTER [PD_MONGO]): INTEGER_32
		require
			not a_conn.is_void
		local
			l_conn: CE_POINTER [PD_MONGO]
		do
			l_conn := a_conn
			Result := (l_conn.item.sock)
		end

	mongo_cursor_create: CE_POINTER [PD_MONGO_CURSOR]
		do
			Result := (create {CE_CLASS_POINTER [PD_MONGO_CURSOR]}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (((create {PD_MONGO_CURSOR}.make).structure_size.to_natural_32).to_integer_32))))
		end

	mongo_cursor_dispose (a_cursor: CE_POINTER [PD_MONGO_CURSOR])
		do
			free (create {CE_VOID_POINTER}.make_cast (a_cursor))
		end

	mongo_ns_to_cmd_db (a_ns: CE_POINTER [INTEGER_8]): CE_POINTER [INTEGER_8]
		require
			not a_ns.is_void
		local
			l_current: CE_POINTER [INTEGER_8]
			l_cmd_db_name: CE_POINTER [INTEGER_8]
			l_len: INTEGER_32
		do
			l_current := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_cmd_db_name := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_len := 0
			l_current := a_ns
			from
			until
				(l_current[0]).to_integer_32 = 46
			loop
				l_len := l_len + 1
				l_current := l_current + 1
			end
			l_cmd_db_name := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (l_len + 6)))
			strncpy (l_cmd_db_name, a_ns, l_len.to_natural_32).do_nothing
			strncpy (l_cmd_db_name + l_len, ce_cstring (".$cmd"), {NATURAL_32} 6).do_nothing
			Result := l_cmd_db_name
		end

	mongo_validate_ns (a_conn: CE_POINTER [PD_MONGO]; a_ns: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_ns.is_void
		local
			l_last: CE_POINTER [INTEGER_8]
			l_current: CE_POINTER [INTEGER_8]
			l_db_name: CE_POINTER [INTEGER_8]
			l_collection_name: CE_POINTER [INTEGER_8]
			l_errmsg: CE_ARRAY [INTEGER_8]
			l_ns_len: INTEGER_32
			l_pos: INTEGER_32
			l_return, l_break: BOOLEAN
		do
			create {CE_INTEGER_8_ARRAY} l_errmsg.make (<<64>>)
			create {CE_INTEGER_8_POINTER} l_db_name.make
			create {CE_INTEGER_8_POINTER} l_current.make
			create {CE_INTEGER_8_POINTER} l_collection_name.make
			create {CE_INTEGER_8_POINTER} l_last.make

			from until l_pos = -1 loop -- -1: return
				inspect l_pos

				-- <function body>
				when 0 then -- execution start:
					l_last := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
					l_current := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
					l_db_name := a_ns
					l_collection_name := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
					l_ns_len := 0
					if (a_ns[0]).to_integer_32 = 46 then
						a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("ns cannot start with a %'.%'."), 0)
						Result := (-1)
						l_return := True
					end
					if not l_return then
						l_current := a_ns
						from
						until
							not (not (l_break) and then ((l_current[0]).to_integer_32 /= 0))
						loop
							l_break := False
							if (l_current[0]).to_integer_32 = 46 then
								l_current := l_current + 1
								l_break := True
							end
							if not (l_break) then
								l_current := l_current + 1
							end
						end
						l_break := False
						if (l_current[0]).to_integer_32 = 0 then
							a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("ns cannot start with a %'.%'."), 0)
							Result := (-1)
							l_return := True
						elseif (l_current[0]).to_integer_32 = 46 then
							a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("ns cannot start with a %'.%'."), 0)
							Result := (-1)
							l_return := True
						end
					end
					if not l_return and then ((l_current[1]).to_integer_32 = 0) then
						a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("Collection name missing."), 0)
						Result := (-1)
						l_return := True
					end
					if not l_return then
						l_collection_name := l_current
						if (l_collection_name - 1).is_equal (l_db_name) then
							a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("Database name missing."), 0)
							Result := (-1)
							l_return := True
						end
					end
					if l_return then
						l_pos := -1 + 7 -- skip if block
					end
				when 1 then -- if:
						l_current := l_db_name
				when 2 then -- loop condition:
						if not (not (l_return) and then ((l_current[0]).to_integer_32 /= 46)) then
							l_pos := -1 + 7 -- skip loop
						end
				when 3 then -- loop body:
						inspect (l_current[0]).to_integer_32 -- switch 1 head
						when 92 then l_pos := -1 + 4
						when 47 then l_pos := -1 + 4
						when 36 then l_pos := -1 + 4
						when 32 then l_pos := -1 + 4
						else l_pos := -1 + 5
						end
				when 4 then -- switch 1, case 92 | switch 1, case 47 | switch 1, case 36 | switch 1, case 32
						a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("Database name may not contain %' %', %'$%', %'/%', or %'\%'"), 0)
						Result := (-1)
						l_return := True
				when 5 then -- switch 1, default:
						l_pos := -1 + 6 -- switch 1; break
				when 6 then -- switch 1, done:
						if not l_return then
							l_ns_len := l_ns_len + 1
							l_current := l_current + 1
						end
						l_pos := -1 + 2 -- back to loop condition
				when 7 then -- loop done | if done
					if not l_return then
						l_ns_len := l_ns_len + 1
						l_current := l_collection_name
						from
						until
							not (not (l_return) and then ((l_current[0]).to_integer_32 /= 0))
						loop
							if (not l_last.is_void) then
								if (l_last[0]).to_integer_32 = 46 then
									if (l_current[0]).to_integer_32 = 46 then
										a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("Collection may not contain two consecutive %'.%'"), 0)
										Result := (-1)
										l_return := True
									end
								end
							end
							if not l_return and then ((l_current[0]).to_integer_32 = 36) then
								a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("Collection may not contain %'$%'"), 0)
								Result := (-1)
								l_return := True
							end
							if not l_return then
								l_last := l_current
								l_ns_len := l_ns_len + 1
								l_current := l_current + 1
							end
						end
					end
					if not l_return and then (l_ns_len > 128) then
						(create {P_SRC_BSON_STATIC}).bson_sprintf.item.call ([l_errmsg, ce_cstring ("Namespace too long; has %%d but must <= 128."), l_ns_len])
						a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, l_errmsg, 0)
						Result := (-1)
						l_return := True
					end
					if not l_return and then ((l_current[-1]).to_integer_32 = 46) then
						a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_ns_invalid, ce_cstring ("Collection may not end with %'.%'"), 0)
						Result := (-1)
						l_return := True
					end
					if not l_return then
						Result := 0
					end
				when 8 then
					l_pos := -1 + -1 -- end of function
				end
				-- </function body>

				l_pos := l_pos + 1
			end
		end

	mongo_message_create (a_len, a_id, a_response_to, a_op: INTEGER_32): CE_POINTER [PD_MONGO_MESSAGE]
		local
			l_id: INTEGER_32
			l_mm: CE_POINTER [PD_MONGO_MESSAGE]
		do
			l_id := a_id
			l_mm := create {CE_CLASS_POINTER [PD_MONGO_MESSAGE]}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (a_len)))
			if not l_id.to_boolean then
				l_id := rand
			end
			l_mm.item.head.len := a_len
			l_mm.item.head.id := l_id
			l_mm.item.head.response_to := a_response_to
			l_mm.item.head.op := a_op
			Result := l_mm
		end

	mongo_message_send (a_conn: CE_POINTER [PD_MONGO]; a_mm: CE_POINTER [PD_MONGO_MESSAGE]): INTEGER_32
		local
			l_head: CE_ARRAY [PD_MONGO_HEADER]
			l_res: INTEGER_32
			l_return: BOOLEAN
		do
			create {CE_CLASS_ARRAY [PD_MONGO_HEADER]} l_head.make (<<1>>)

			memcpy (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (l_head.item.c_pointer + (create {PD_MONGO_HEADER}.make).offset_of ("len")))), create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (a_mm.c_pointer + (create {PD_MONGO_MESSAGE}.make).offset_of ("head") + (create {PD_MONGO_HEADER}.make).offset_of ("len")))), {NATURAL_32} 4).do_nothing
			memcpy (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (l_head.item.c_pointer + (create {PD_MONGO_HEADER}.make).offset_of ("id")))), create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (a_mm.c_pointer + (create {PD_MONGO_MESSAGE}.make).offset_of ("head") + (create {PD_MONGO_HEADER}.make).offset_of ("id")))), {NATURAL_32} 4).do_nothing
			memcpy (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (l_head.item.c_pointer + (create {PD_MONGO_HEADER}.make).offset_of ("response_to")))), create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (a_mm.c_pointer + (create {PD_MONGO_MESSAGE}.make).offset_of ("head") + (create {PD_MONGO_HEADER}.make).offset_of ("response_to")))), {NATURAL_32} 4).do_nothing
			memcpy (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (l_head.item.c_pointer + (create {PD_MONGO_HEADER}.make).offset_of ("op")))), create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (a_mm.c_pointer + (create {PD_MONGO_MESSAGE}.make).offset_of ("head") + (create {PD_MONGO_HEADER}.make).offset_of ("op")))), {NATURAL_32} 4).do_nothing
			l_res := (create {P_SRC_ENV_STATIC}).env_write_socket (a_conn, create {CE_VOID_POINTER}.make_cast (l_head), ((create {PD_MONGO_HEADER}.make).structure_size.to_natural_32).to_integer_32)
			if l_res /= 0 then
				(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (a_mm))
				Result := l_res
				l_return := True
			end
			if not l_return then
				l_res := (create {P_SRC_ENV_STATIC}).env_write_socket (a_conn, create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_8_POINTER}.make_shared (a_mm.c_pointer + (create {PD_MONGO_MESSAGE}.make).offset_of ("data")))), ((a_mm.item.head.len).to_natural_32 - (create {PD_MONGO_HEADER}.make).structure_size.to_natural_32).to_integer_32)
				if l_res /= 0 then
					(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (a_mm))
					Result := l_res
					l_return := True
				end
			end
			if not l_return then
				(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (a_mm))
				Result := 0
			end
		end

	mongo_data_append (a_start: CE_POINTER [INTEGER_8]; a_data_0: CE_POINTER [ANY]; a_len: INTEGER_32): CE_POINTER [INTEGER_8]
		do
			memcpy (create {CE_VOID_POINTER}.make_cast (a_start), a_data_0, a_len.to_natural_32).do_nothing
			Result := (a_start + a_len)
		end

	mongo_data_append32 (a_start: CE_POINTER [INTEGER_8]; a_data_0: CE_POINTER [ANY]): CE_POINTER [INTEGER_8]
		do
			memcpy (create {CE_VOID_POINTER}.make_cast (a_start), a_data_0, {NATURAL_32} 4).do_nothing
			Result := (a_start + 4)
		end

	mongo_data_append64 (a_start: CE_POINTER [INTEGER_8]; a_data_0: CE_POINTER [ANY]): CE_POINTER [INTEGER_8]
		do
			memcpy (create {CE_VOID_POINTER}.make_cast (a_start), a_data_0, (8).to_natural_32).do_nothing
			Result := (a_start + 8)
		end

	mongo_check_last_error (a_conn: CE_POINTER [PD_MONGO]; a_ns: CE_POINTER [INTEGER_8]; a_write_concern: CE_POINTER [PD_MONGO_WRITE_CONCERN]): INTEGER_32
		require
			not a_ns.is_void
			not a_write_concern.is_void
		local
			l_response: CE_ARRAY [PD_BSON]
			l_fields: CE_ARRAY [PD_BSON]
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_res: INTEGER_32
			l_cmd_ns: CE_POINTER [INTEGER_8]
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_response.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_fields.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_response.item.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_response.item.cur := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_response.item.data_size := 0
			l_response.item.finished := 0
			l_response.item.stack[0] := 0
			l_response.item.stack[1] := 0
			l_response.item.stack[2] := 0
			l_response.item.stack[3] := 0
			l_response.item.stack[4] := 0
			l_response.item.stack[5] := 0
			l_response.item.stack[6] := 0
			l_response.item.stack[7] := 0
			l_response.item.stack[8] := 0
			l_response.item.stack[9] := 0
			l_response.item.stack[10] := 0
			l_response.item.stack[11] := 0
			l_response.item.stack[12] := 0
			l_response.item.stack[13] := 0
			l_response.item.stack[14] := 0
			l_response.item.stack[15] := 0
			l_response.item.stack[16] := 0
			l_response.item.stack[17] := 0
			l_response.item.stack[18] := 0
			l_response.item.stack[19] := 0
			l_response.item.stack[20] := 0
			l_response.item.stack[21] := 0
			l_response.item.stack[22] := 0
			l_response.item.stack[23] := 0
			l_response.item.stack[24] := 0
			l_response.item.stack[25] := 0
			l_response.item.stack[26] := 0
			l_response.item.stack[27] := 0
			l_response.item.stack[28] := 0
			l_response.item.stack[29] := 0
			l_response.item.stack[30] := 0
			l_response.item.stack[31] := 0
			l_response.item.stack_pos := 0
			l_response.item.err := 0
			l_response.item.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_res := 0
			l_cmd_ns := (mongo_ns_to_cmd_db (a_ns))
			l_res := mongo_find_one (a_conn, l_cmd_ns, a_write_concern.item.cmd, (l_fields.item.empty), l_response);
			(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (l_cmd_ns))
			if l_res /= 0 then
				Result := (-1)
			elseif (l_it.item.bson_find (l_response, ce_cstring ("$err"))) = {NATURAL_32} 2 then
				a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_write_error, ce_cstring ("See conn->lasterrstr for details."), 0)
				a_conn.item.set_last_error (l_it, l_response)
				Result := (-1)
			elseif (l_it.item.bson_find (l_response, ce_cstring ("err"))) = {NATURAL_32} 2 then
				a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_write_error, ce_cstring ("See conn->lasterrstr for details."), 0)
				a_conn.item.set_last_error (l_it, l_response)
				Result := (-1)
			else
				Result := 0
			end
		end

	mongo_choose_write_concern (a_conn: CE_POINTER [PD_MONGO]; a_custom_write_concern: CE_POINTER [PD_MONGO_WRITE_CONCERN]; a_write_concern: CE_POINTER [CE_POINTER [PD_MONGO_WRITE_CONCERN]]): INTEGER_32
		require
			not a_write_concern.is_void
		do
			if (not a_custom_write_concern.is_void) then
				a_write_concern.item := a_custom_write_concern
			elseif (not (a_conn.item.write_concern).is_void) then
				a_write_concern.item := a_conn.item.write_concern
			end
			if (not (a_write_concern.item).is_void) then
				if ((a_write_concern.item).item.cmd).is_void then
					a_conn.item.set_error ({PC_MONGO_ERROR_T}.Mongo_write_concern_invalid, ce_cstring ("Must call mongo_write_concern_finish() before using *write_concern."), 0)
					Result := (-1)
				else
					Result := 0
				end
			else
				Result := 0
			end
		end

	mongo_insert_batch (a_conn: CE_POINTER [PD_MONGO]; a_ns: CE_POINTER [INTEGER_8]; a_bsons: CE_POINTER [CE_POINTER [PD_BSON]]; a_count: INTEGER_32; a_custom_write_concern: CE_POINTER [PD_MONGO_WRITE_CONCERN]; a_flags: INTEGER_32): INTEGER_32
		require
			not a_ns.is_void
		local
			l_mm: CE_POINTER [PD_MONGO_MESSAGE]
			l_write_concern: CE_ARRAY [CE_POINTER [PD_MONGO_WRITE_CONCERN]]
			l_i: INTEGER_32
			l_data_0: CE_POINTER [INTEGER_8]
			l_overhead: INTEGER_32
			l_size: INTEGER_32
			l_return: BOOLEAN
		do
			create {CE_POINTER_ARRAY [PD_MONGO_WRITE_CONCERN]} l_write_concern.make (<<1>>)

			l_write_concern.item := create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer)
			l_overhead := (({NATURAL_32} 20 + (strlen (a_ns))) + {NATURAL_32} 1).to_integer_32
			l_size := l_overhead
			if (mongo_validate_ns (a_conn, a_ns)) /= 0 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				l_i := 0
				from
				until
					not (not (l_return) and then (l_i < a_count))
				loop
					l_size := l_size + ((create {P_SRC_BSON_STATIC}).bson_size (a_bsons[l_i]))
					if (a_conn.item.bson_valid (a_bsons[l_i], 1)) /= 0 then
						Result := (-1)
						l_return := True
					end
					if not l_return then
						l_i := l_i + 1
					end
				end
			end
			if not l_return and then (l_size - l_overhead > a_conn.item.max_bson_size) then
				a_conn.item.err := {PC_MONGO_ERROR_T}.Mongo_bson_too_large
				Result := (-1)
				l_return := True
			end
			if not l_return then
				if (mongo_choose_write_concern (a_conn, a_custom_write_concern, l_write_concern)) = -1 then
					Result := (-1)
					l_return := True
				end
			end
			if not l_return then
				l_mm := mongo_message_create (l_size, 0, 0, 2_002)
				l_data_0 := create {CE_INTEGER_8_POINTER}.make_shared (l_mm.c_pointer + (create {PD_MONGO_MESSAGE}.make).offset_of ("data"))
				if ((a_flags).bit_and (1)).to_boolean then
					l_data_0 := mongo_data_append32 (l_data_0, create {CE_VOID_POINTER}.make_cast (one.item))
				else
					l_data_0 := mongo_data_append32 (l_data_0, create {CE_VOID_POINTER}.make_cast (zero.item))
				end
				l_data_0 := mongo_data_append (l_data_0, create {CE_VOID_POINTER}.make_cast (a_ns), ((strlen (a_ns)) + {NATURAL_32} 1).to_integer_32)
				l_i := 0
				from
				until
					not (l_i < a_count)
				loop
					l_data_0 := mongo_data_append (l_data_0, create {CE_VOID_POINTER}.make_cast ((a_bsons[l_i]).item.data), ((create {P_SRC_BSON_STATIC}).bson_size (a_bsons[l_i])))
					l_i := l_i + 1
				end
				if (not (l_write_concern.item).is_void) then
					if (mongo_message_send (a_conn, l_mm)) = -1 then
						Result := (-1)
						l_return := True
					end
					if not l_return then
						Result := (mongo_check_last_error (a_conn, a_ns, l_write_concern.item))
					end
				else
					Result := (mongo_message_send (a_conn, l_mm))
				end
			end
		end

	mongo_write_concern_init (a_write_concern: CE_POINTER [PD_MONGO_WRITE_CONCERN])
		do
			memset (create {CE_VOID_POINTER}.make_cast (a_write_concern), 0, (create {PD_MONGO_WRITE_CONCERN}.make).structure_size.to_natural_32).do_nothing
		end

	mongo_write_concern_destroy (a_write_concern: CE_POINTER [PD_MONGO_WRITE_CONCERN])
		local
			l_return: BOOLEAN
		do
			if a_write_concern.is_void then
				l_return := True
			end
			if not l_return then
				if (not (a_write_concern.item.cmd).is_void) then
					a_write_concern.item.cmd.item.destroy
				end
				(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (a_write_concern.item.cmd))
			end
		end

	mongo_find (a_conn: CE_POINTER [PD_MONGO]; a_ns: CE_POINTER [INTEGER_8]; a_query, a_fields: CE_POINTER [PD_BSON]; a_limit, a_skip, a_options: INTEGER_32): CE_POINTER [PD_MONGO_CURSOR]
		local
			l_cursor: CE_POINTER [PD_MONGO_CURSOR]
		do
			l_cursor := create {CE_CLASS_POINTER [PD_MONGO_CURSOR]}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (((create {PD_MONGO_CURSOR}.make).structure_size.to_natural_32).to_integer_32)))
			l_cursor.item.init (a_conn, a_ns)
			l_cursor.item.flags := (l_cursor.item.flags).bit_or ((1))
			l_cursor.item.set_query (a_query)
			l_cursor.item.set_fields (a_fields)
			l_cursor.item.set_limit (a_limit)
			l_cursor.item.set_skip (a_skip)
			l_cursor.item.set_options (a_options)
			if (l_cursor.item.op_query) = 0 then
				Result := l_cursor
			else
				l_cursor.item.destroy.do_nothing
				Result := (create {CE_CLASS_POINTER [PD_MONGO_CURSOR]}.make_shared (default_pointer))
			end
		end

	mongo_find_one (a_conn: CE_POINTER [PD_MONGO]; a_ns: CE_POINTER [INTEGER_8]; a_query, a_fields, a_out: CE_POINTER [PD_BSON]): INTEGER_32
		local
			l_cursor: CE_ARRAY [PD_MONGO_CURSOR]
		do
			create {CE_CLASS_ARRAY [PD_MONGO_CURSOR]} l_cursor.make (<<1>>)

			l_cursor.item.init (a_conn, a_ns)
			l_cursor.item.set_query (a_query)
			l_cursor.item.set_fields (a_fields)
			l_cursor.item.set_limit (1)
			if (l_cursor.item.next) = 0 then
				if (not a_out.is_void) then
					a_out.item.init_size2 (((create {P_SRC_BSON_STATIC}).bson_size ((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (l_cursor.c_pointer + (create {PD_MONGO_CURSOR}.make).structure_size * (0) + (create {PD_MONGO_CURSOR}.make).offset_of ("current2"))))))
					memcpy (create {CE_VOID_POINTER}.make_cast (a_out.item.data), create {CE_VOID_POINTER}.make_cast (l_cursor[0].current2.data), ((create {P_SRC_BSON_STATIC}).bson_size ((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (l_cursor.c_pointer + (create {PD_MONGO_CURSOR}.make).structure_size * (0) + (create {PD_MONGO_CURSOR}.make).offset_of ("current2"))))).to_natural_32).do_nothing
					a_out.item.finished := 1
				end
				l_cursor.item.destroy.do_nothing
				Result := 0
			else
				l_cursor.item.destroy.do_nothing
				Result := (-1)
			end
		end

	mongo_cursor_bson (a_cursor: CE_POINTER [PD_MONGO_CURSOR]): CE_POINTER [PD_BSON]
		do
			Result := (create {CE_CLASS_POINTER [PD_BSON]}.make_shared (a_cursor.c_pointer + (create {PD_MONGO_CURSOR}.make).offset_of ("current2")))
		end

	mongo_create_capped_collection (a_conn: CE_POINTER [PD_MONGO]; a_db, a_collection: CE_POINTER [INTEGER_8]; a_size, a_max: INTEGER_32; a_out: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_collection.is_void
		local
			l_b: CE_ARRAY [PD_BSON]
			l_result: INTEGER_32
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_b.make (<<1>>)

			l_b.item.init;
			l_b.item.append_string (ce_cstring ("create"), a_collection).do_nothing
			l_b.item.append_bool (ce_cstring ("capped"), 1).do_nothing
			l_b.item.append_int (ce_cstring ("size"), a_size).do_nothing
			if a_max > 0 then
				l_b.item.append_int (ce_cstring ("max"), a_size).do_nothing
			end
			l_b.item.finish.do_nothing
			l_result := mongo_run_command (a_conn, a_db, l_b, a_out)
			l_b.item.destroy
			Result := l_result
		end

	mongo_count (a_conn: CE_POINTER [PD_MONGO]; a_db, a_ns: CE_POINTER [INTEGER_8]; a_query: CE_POINTER [PD_BSON]): REAL_64
		require
			not a_ns.is_void
		local
			l_cmd: CE_ARRAY [PD_BSON]
			l_out: CE_ARRAY [PD_BSON]
			l_count: REAL_64
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_cmd.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_out.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_out.item.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.cur := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.data_size := 0
			l_out.item.finished := 0
			l_out.item.stack[0] := 0
			l_out.item.stack[1] := 0
			l_out.item.stack[2] := 0
			l_out.item.stack[3] := 0
			l_out.item.stack[4] := 0
			l_out.item.stack[5] := 0
			l_out.item.stack[6] := 0
			l_out.item.stack[7] := 0
			l_out.item.stack[8] := 0
			l_out.item.stack[9] := 0
			l_out.item.stack[10] := 0
			l_out.item.stack[11] := 0
			l_out.item.stack[12] := 0
			l_out.item.stack[13] := 0
			l_out.item.stack[14] := 0
			l_out.item.stack[15] := 0
			l_out.item.stack[16] := 0
			l_out.item.stack[17] := 0
			l_out.item.stack[18] := 0
			l_out.item.stack[19] := 0
			l_out.item.stack[20] := 0
			l_out.item.stack[21] := 0
			l_out.item.stack[22] := 0
			l_out.item.stack[23] := 0
			l_out.item.stack[24] := 0
			l_out.item.stack[25] := 0
			l_out.item.stack[26] := 0
			l_out.item.stack[27] := 0
			l_out.item.stack[28] := 0
			l_out.item.stack[29] := 0
			l_out.item.stack[30] := 0
			l_out.item.stack[31] := 0
			l_out.item.stack_pos := 0
			l_out.item.err := 0
			l_out.item.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_count := (-1).to_double
			l_cmd.item.init;
			l_cmd.item.append_string (ce_cstring ("count"), a_ns).do_nothing
			if (not a_query.is_void) then
				if ((create {P_SRC_BSON_STATIC}).bson_size (a_query)) > 5 then
					l_cmd.item.append_bson (ce_cstring ("query"), a_query).do_nothing
				end
			end
			l_cmd.item.finish.do_nothing
			if (mongo_run_command (a_conn, a_db, l_cmd, l_out)) = 0 then
				if (l_it.item.bson_find (l_out, ce_cstring ("n"))).to_boolean then
					l_count := l_it.item.double
				end
				l_cmd.item.destroy
				l_out.item.destroy
				Result := l_count
			else
				l_out.item.destroy
				l_cmd.item.destroy
				Result := (-1).to_double
			end
		end

	mongo_run_command (a_conn: CE_POINTER [PD_MONGO]; a_db: CE_POINTER [INTEGER_8]; a_command, a_out: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_db.is_void
		local
			l_response: CE_ARRAY [PD_BSON]
			l_fields: CE_ARRAY [PD_BSON]
			l_sl: INTEGER_32
			l_ns: CE_POINTER [INTEGER_8]
			l_res: INTEGER_32
			l_success: INTEGER_32
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_response.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_fields.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_response.item.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_response.item.cur := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_response.item.data_size := 0
			l_response.item.finished := 0
			l_response.item.stack[0] := 0
			l_response.item.stack[1] := 0
			l_response.item.stack[2] := 0
			l_response.item.stack[3] := 0
			l_response.item.stack[4] := 0
			l_response.item.stack[5] := 0
			l_response.item.stack[6] := 0
			l_response.item.stack[7] := 0
			l_response.item.stack[8] := 0
			l_response.item.stack[9] := 0
			l_response.item.stack[10] := 0
			l_response.item.stack[11] := 0
			l_response.item.stack[12] := 0
			l_response.item.stack[13] := 0
			l_response.item.stack[14] := 0
			l_response.item.stack[15] := 0
			l_response.item.stack[16] := 0
			l_response.item.stack[17] := 0
			l_response.item.stack[18] := 0
			l_response.item.stack[19] := 0
			l_response.item.stack[20] := 0
			l_response.item.stack[21] := 0
			l_response.item.stack[22] := 0
			l_response.item.stack[23] := 0
			l_response.item.stack[24] := 0
			l_response.item.stack[25] := 0
			l_response.item.stack[26] := 0
			l_response.item.stack[27] := 0
			l_response.item.stack[28] := 0
			l_response.item.stack[29] := 0
			l_response.item.stack[30] := 0
			l_response.item.stack[31] := 0
			l_response.item.stack_pos := 0
			l_response.item.err := 0
			l_response.item.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_sl := (strlen (a_db)).to_integer_32
			l_ns := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc ((l_sl + 5) + 1)))
			l_success := 0
			strcpy (l_ns, a_db).do_nothing
			strcpy (l_ns + l_sl, ce_cstring (".$cmd")).do_nothing
			l_res := mongo_find_one (a_conn, l_ns, a_command, (l_fields.item.empty), l_response);
			(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (l_ns))
			if l_res /= 0 then
				Result := (-1)
			else
				if (l_it.item.bson_find (l_response, ce_cstring ("ok"))).to_boolean then
					l_success := l_it.item.bool
				end
				if not l_success.to_boolean then
					a_conn.item.err := {PC_MONGO_ERROR_T}.Mongo_command_failed
					Result := (-1)
				else
					if (not a_out.is_void) then
						a_out.item := (l_response.item).twin
					end
					Result := 0
				end
			end
		end

	mongo_simple_int_command (a_conn: CE_POINTER [PD_MONGO]; a_db, a_cmdstr: CE_POINTER [INTEGER_8]; a_arg: INTEGER_32; a_realout: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_db.is_void
			not a_cmdstr.is_void
		local
			l_out: CE_ARRAY [PD_BSON]
			l_cmd: CE_ARRAY [PD_BSON]
			l_result: INTEGER_32
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_out.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_cmd.make (<<1>>)

			l_out.item.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.cur := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.data_size := 0
			l_out.item.finished := 0
			l_out.item.stack[0] := 0
			l_out.item.stack[1] := 0
			l_out.item.stack[2] := 0
			l_out.item.stack[3] := 0
			l_out.item.stack[4] := 0
			l_out.item.stack[5] := 0
			l_out.item.stack[6] := 0
			l_out.item.stack[7] := 0
			l_out.item.stack[8] := 0
			l_out.item.stack[9] := 0
			l_out.item.stack[10] := 0
			l_out.item.stack[11] := 0
			l_out.item.stack[12] := 0
			l_out.item.stack[13] := 0
			l_out.item.stack[14] := 0
			l_out.item.stack[15] := 0
			l_out.item.stack[16] := 0
			l_out.item.stack[17] := 0
			l_out.item.stack[18] := 0
			l_out.item.stack[19] := 0
			l_out.item.stack[20] := 0
			l_out.item.stack[21] := 0
			l_out.item.stack[22] := 0
			l_out.item.stack[23] := 0
			l_out.item.stack[24] := 0
			l_out.item.stack[25] := 0
			l_out.item.stack[26] := 0
			l_out.item.stack[27] := 0
			l_out.item.stack[28] := 0
			l_out.item.stack[29] := 0
			l_out.item.stack[30] := 0
			l_out.item.stack[31] := 0
			l_out.item.stack_pos := 0
			l_out.item.err := 0
			l_out.item.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_cmd.item.init
			l_cmd.item.append_int (a_cmdstr, a_arg).do_nothing
			l_cmd.item.finish.do_nothing
			l_result := mongo_run_command (a_conn, a_db, l_cmd, l_out);
			l_cmd.item.destroy
			if (not a_realout.is_void) then
				a_realout.item := (l_out.item).twin
			else
				l_out.item.destroy
			end
			Result := l_result
		end

	mongo_simple_str_command (a_conn: CE_POINTER [PD_MONGO]; a_db, a_cmdstr, a_arg: CE_POINTER [INTEGER_8]; a_realout: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_db.is_void
			not a_arg.is_void
		local
			l_out: CE_ARRAY [PD_BSON]
			l_result: INTEGER_32
			l_cmd: CE_ARRAY [PD_BSON]
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_out.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_cmd.make (<<1>>)

			l_out.item.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.cur := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.data_size := 0
			l_out.item.finished := 0
			l_out.item.stack[0] := 0
			l_out.item.stack[1] := 0
			l_out.item.stack[2] := 0
			l_out.item.stack[3] := 0
			l_out.item.stack[4] := 0
			l_out.item.stack[5] := 0
			l_out.item.stack[6] := 0
			l_out.item.stack[7] := 0
			l_out.item.stack[8] := 0
			l_out.item.stack[9] := 0
			l_out.item.stack[10] := 0
			l_out.item.stack[11] := 0
			l_out.item.stack[12] := 0
			l_out.item.stack[13] := 0
			l_out.item.stack[14] := 0
			l_out.item.stack[15] := 0
			l_out.item.stack[16] := 0
			l_out.item.stack[17] := 0
			l_out.item.stack[18] := 0
			l_out.item.stack[19] := 0
			l_out.item.stack[20] := 0
			l_out.item.stack[21] := 0
			l_out.item.stack[22] := 0
			l_out.item.stack[23] := 0
			l_out.item.stack[24] := 0
			l_out.item.stack[25] := 0
			l_out.item.stack[26] := 0
			l_out.item.stack[27] := 0
			l_out.item.stack[28] := 0
			l_out.item.stack[29] := 0
			l_out.item.stack[30] := 0
			l_out.item.stack[31] := 0
			l_out.item.stack_pos := 0
			l_out.item.err := 0
			l_out.item.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_cmd.item.init;
			l_cmd.item.append_string (a_cmdstr, a_arg).do_nothing
			l_cmd.item.finish.do_nothing
			l_result := mongo_run_command (a_conn, a_db, l_cmd, l_out);
			l_cmd.item.destroy
			if (not a_realout.is_void) then
				a_realout.item := (l_out.item).twin
			else
				l_out.item.destroy
			end
			Result := l_result
		end

	mongo_cmd_drop_db (a_conn: CE_POINTER [PD_MONGO]; a_db: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_db.is_void
		do
			Result := (mongo_simple_int_command (a_conn, a_db, ce_cstring ("dropDatabase"), 1, create {CE_CLASS_POINTER [PD_BSON]}.make_shared (default_pointer)))
		end

	mongo_cmd_drop_collection (a_conn: CE_POINTER [PD_MONGO]; a_db, a_collection: CE_POINTER [INTEGER_8]; a_out: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_db.is_void
			not a_collection.is_void
		do
			Result := (mongo_simple_str_command (a_conn, a_db, ce_cstring ("drop"), a_collection, a_out))
		end

	mongo_cmd_reset_error (a_conn: CE_POINTER [PD_MONGO]; a_db: CE_POINTER [INTEGER_8])
		require
			not a_db.is_void
		do
			mongo_simple_int_command (a_conn, a_db, ce_cstring ("reseterror"), 1, create {CE_CLASS_POINTER [PD_BSON]}.make_shared (default_pointer)).do_nothing
		end

	mongo_cmd_ismaster (a_conn: CE_POINTER [PD_MONGO]; a_realout: CE_POINTER [PD_BSON]): INTEGER_32
		local
			l_out: CE_ARRAY [PD_BSON]
			l_ismaster: INTEGER_32
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_out.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_out.item.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.cur := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_out.item.data_size := 0
			l_out.item.finished := 0
			l_out.item.stack[0] := 0
			l_out.item.stack[1] := 0
			l_out.item.stack[2] := 0
			l_out.item.stack[3] := 0
			l_out.item.stack[4] := 0
			l_out.item.stack[5] := 0
			l_out.item.stack[6] := 0
			l_out.item.stack[7] := 0
			l_out.item.stack[8] := 0
			l_out.item.stack[9] := 0
			l_out.item.stack[10] := 0
			l_out.item.stack[11] := 0
			l_out.item.stack[12] := 0
			l_out.item.stack[13] := 0
			l_out.item.stack[14] := 0
			l_out.item.stack[15] := 0
			l_out.item.stack[16] := 0
			l_out.item.stack[17] := 0
			l_out.item.stack[18] := 0
			l_out.item.stack[19] := 0
			l_out.item.stack[20] := 0
			l_out.item.stack[21] := 0
			l_out.item.stack[22] := 0
			l_out.item.stack[23] := 0
			l_out.item.stack[24] := 0
			l_out.item.stack[25] := 0
			l_out.item.stack[26] := 0
			l_out.item.stack[27] := 0
			l_out.item.stack[28] := 0
			l_out.item.stack[29] := 0
			l_out.item.stack[30] := 0
			l_out.item.stack[31] := 0
			l_out.item.stack_pos := 0
			l_out.item.err := 0
			l_out.item.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			l_ismaster := 0
			if (mongo_simple_int_command (a_conn, ce_cstring ("admin"), ce_cstring ("ismaster"), 1, l_out)) = 0 then
				l_it.item.bson_find (l_out, ce_cstring ("ismaster")).do_nothing
				l_ismaster := l_it.item.bool
			end
			if (not a_realout.is_void) then
				a_realout.item := (l_out.item).twin
			else
				l_out.item.destroy
			end
			Result := l_ismaster
		end

	digest2hex (a_digest: CE_POINTER [NATURAL_8]; a_hex_digest: CE_POINTER [INTEGER_8])
		require
			not a_hex_digest.is_void
		local
			l_i: INTEGER_32
		do
			l_i := 0
			from
			until
				not (l_i < 16)
			loop
				a_hex_digest[2 * l_i] := hex.item[((((a_digest[l_i]).to_integer_32).bit_and (240))).bit_shift_right (4)]
				a_hex_digest[(2 * l_i + 1)] := hex.item[((a_digest[l_i]).to_integer_32).bit_and (15)]
				l_i := l_i + 1
			end
			a_hex_digest[32] := (('%/0/').code).to_integer_8
		end

	mongo_pass_digest (a_user, a_pass, a_hex_digest: CE_POINTER [INTEGER_8])
		require
			not a_user.is_void
			not a_pass.is_void
			not a_hex_digest.is_void
		local
			l_st: CE_ARRAY [PD_MONGO_MD5_STATE_S]
			l_digest: CE_ARRAY [NATURAL_8]
		do
			create {CE_CLASS_ARRAY [PD_MONGO_MD5_STATE_S]} l_st.make (<<1>>)
			create {CE_NATURAL_8_ARRAY} l_digest.make (<<16>>)

			l_st.item.mongo_md5_init
			l_st.item.mongo_md5_append (create {CE_NATURAL_8_POINTER}.make_cast (a_user), (strlen (a_user)).to_integer_32)
			l_st.item.mongo_md5_append (create {CE_NATURAL_8_POINTER}.make_cast (ce_cstring (":mongo:")), 7)
			l_st.item.mongo_md5_append (create {CE_NATURAL_8_POINTER}.make_cast (a_pass), (strlen (a_pass)).to_integer_32)
			l_st.item.mongo_md5_finish (create {CE_NATURAL_8_POINTER}.make_cast (l_digest))
			digest2hex (create {CE_NATURAL_8_POINTER}.make_cast (l_digest), a_hex_digest)
		end

	mongo_cmd_authenticate (a_conn: CE_POINTER [PD_MONGO]; a_db, a_user, a_pass: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_db.is_void
		local
			l_from_db: CE_ARRAY [PD_BSON]
			l_cmd: CE_ARRAY [PD_BSON]
			l_out: CE_ARRAY [PD_BSON]
			l_nonce: CE_POINTER [INTEGER_8]
			l_result: INTEGER_32
			l_st: CE_ARRAY [PD_MONGO_MD5_STATE_S]
			l_digest: CE_ARRAY [NATURAL_8]
			l_hex_digest: CE_ARRAY [INTEGER_8]
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_return: BOOLEAN
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_from_db.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_cmd.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_out.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_MONGO_MD5_STATE_S]} l_st.make (<<1>>)
			create {CE_NATURAL_8_ARRAY} l_digest.make (<<16>>)
			create {CE_INTEGER_8_ARRAY} l_hex_digest.make (<<33>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)
			create {CE_INTEGER_8_POINTER} l_nonce.make

			if (mongo_simple_int_command (a_conn, a_db, ce_cstring ("getnonce"), 1, l_from_db)) = 0 then
				l_it.item.bson_find (l_from_db, ce_cstring ("nonce")).do_nothing
				l_nonce := l_it.item.string
			else
				Result := (-1)
				l_return := True
			end
			if not l_return then
				mongo_pass_digest (a_user, a_pass, l_hex_digest)
				l_st.item.mongo_md5_init
				l_st.item.mongo_md5_append (create {CE_NATURAL_8_POINTER}.make_cast (l_nonce), (strlen (l_nonce)).to_integer_32)
				l_st.item.mongo_md5_append (create {CE_NATURAL_8_POINTER}.make_cast (a_user), (strlen (a_user)).to_integer_32)
				l_st.item.mongo_md5_append (create {CE_NATURAL_8_POINTER}.make_cast (l_hex_digest), 32)
				l_st.item.mongo_md5_finish (create {CE_NATURAL_8_POINTER}.make_cast (l_digest))
				digest2hex (create {CE_NATURAL_8_POINTER}.make_cast (l_digest), l_hex_digest)
				l_cmd.item.init
				l_cmd.item.append_int (ce_cstring ("authenticate"), 1).do_nothing;
				l_cmd.item.append_string (ce_cstring ("user"), a_user).do_nothing;
				l_cmd.item.append_string (ce_cstring ("nonce"), l_nonce).do_nothing;
				l_cmd.item.append_string (ce_cstring ("key"), l_hex_digest).do_nothing
				l_cmd.item.finish.do_nothing;
				l_from_db.item.destroy
				l_result := mongo_run_command (a_conn, a_db, l_cmd, l_out)
				l_from_db.item.destroy
				l_cmd.item.destroy
				Result := l_result
			end
		end

end
