class
	PD_BSON

inherit
	CE_UTILITIES
		redefine
			copy,
			is_equal
		end
	CE_CLASS
		redefine
			make,
			make_shared,
			structure_size,
			offset_of,
			copy,
			is_equal
		end
	S_STRING_STATIC
		redefine
			copy,
			is_equal
		end
	S_STDIO_STATIC
		redefine
			copy,
			is_equal
		end
	P_SRC_BSON_STATIC
		redefine
			copy,
			is_equal
		end

create
	make,
	make_shared,
	make_initialized

feature {NONE}

	make
		do
			internal_c_pointer := internal_c_pointer.memory_alloc (structure_size)
			is_allocated_memory := True
		end

	make_shared (a_c_pointer: POINTER)
		do
			internal_c_pointer := a_c_pointer
		end

	make_initialized (a_data: TUPLE [
				a_data: CE_POINTER [INTEGER_8]
				a_cur: CE_POINTER [INTEGER_8]
				a_data_size: INTEGER_32
				a_finished: INTEGER_32
				a_stack: CE_ARRAY [INTEGER_32]
				a_stack_pos: INTEGER_32
				a_err: INTEGER_32
				a_errstr: CE_POINTER [INTEGER_8]
			])
		require else
			a_data /= Void
		do
			make
			Current.data := a_data.a_data
			Current.cur := a_data.a_cur
			Current.data_size := a_data.a_data_size
			Current.finished := a_data.a_finished
			stack.c_pointer.memory_copy (a_data.a_stack.c_pointer, a_data.a_stack.total_count * a_data.a_stack.element_size)
			Current.stack_pos := a_data.a_stack_pos
			Current.err := a_data.a_err
			Current.errstr := a_data.a_errstr
		end

feature {ANY} -- fields

	data: CE_POINTER [INTEGER_8] assign set__data
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__data (internal_c_pointer))
		end

	cur: CE_POINTER [INTEGER_8] assign set__cur
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__cur (internal_c_pointer))
		end

	data_size: INTEGER_32 assign set__data_size
		do
			Result := c__data_size (internal_c_pointer)
		end

	finished: INTEGER_32 assign set__finished
		do
			Result := c__finished (internal_c_pointer)
		end

	stack: CE_ARRAY [INTEGER_32]
		once ("OBJECT")
			create {CE_INTEGER_32_ARRAY} Result.make_shared (c__stack (internal_c_pointer), <<32>>)
		end

	stack_pos: INTEGER_32 assign set__stack_pos
		do
			Result := c__stack_pos (internal_c_pointer)
		end

	err: INTEGER_32 assign set__err
		do
			Result := c__err (internal_c_pointer)
		end

	errstr: CE_POINTER [INTEGER_8] assign set__errstr
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__errstr (internal_c_pointer))
		end

feature {ANY} -- field setters

	set__data (a_data: CE_POINTER [INTEGER_8])
		do
			c__set_data (internal_c_pointer, a_data.c_pointer)
		end

	set__cur (a_cur: CE_POINTER [INTEGER_8])
		do
			c__set_cur (internal_c_pointer, a_cur.c_pointer)
		end

	set__data_size (a_data_size: INTEGER_32)
		do
			c__set_data_size (internal_c_pointer, a_data_size)
		end

	set__finished (a_finished: INTEGER_32)
		do
			c__set_finished (internal_c_pointer, a_finished)
		end

	set__stack_pos (a_stack_pos: INTEGER_32)
		do
			c__set_stack_pos (internal_c_pointer, a_stack_pos)
		end

	set__err (a_err: INTEGER_32)
		do
			c__set_err (internal_c_pointer, a_err)
		end

	set__errstr (a_errstr: CE_POINTER [INTEGER_8])
		do
			c__set_errstr (internal_c_pointer, a_errstr.c_pointer)
		end

feature {PD_BSON} -- private attributes

	zero: CE_STATIC [CE_ARRAY [INTEGER_32]]
		once ("PROCESS")
			create Result.make (create {CE_INTEGER_32_ARRAY}.make_initialized (<<
					0
				>>))
		end

	initial_buffer_size: CE_STATIC [INTEGER_32]
		once ("PROCESS")
			create Result.make (128)
		end

	data3: CE_STATIC [CE_POINTER [INTEGER_8]]
		once ("PROCESS")
			create Result.make (ce_cstring ("%/5/%/0/%/0/%/0/%/0/"))
		end

	printf2: CE_STATIC [CE_ROUTINE]
		once ("PROCESS")
			create Result.make (printf_address)
		end

	hex: CE_STATIC [CE_ARRAY [INTEGER_8]]
		once ("PROCESS")
			create Result.make (create {CE_INTEGER_8_ARRAY}.make (<<16>>))
			across "0123456789abcdef" as c loop
				Result.item[c.target_index - 1] := c.item.code.to_integer_8
			end
		end

feature {ANY} -- public routines

	empty: CE_POINTER [PD_BSON]
		do
			((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.init_data (data3.item).do_nothing
			Current.finished := 1
			Current.err := 0
			Current.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			Current.stack_pos := 0
			Result := ((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer)))
		end

	init_data (a_data_0: CE_POINTER [INTEGER_8]): INTEGER_32
		do
			Current.data := a_data_0
			Result := 0
		end

	reset
		do
			Current.finished := 0
			Current.stack_pos := 0
			Current.err := 0
			Current.errstr := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
		end

	buffer_size: INTEGER_32
		do
			Result := ((cur.distance (data)) + 1)
		end

	data2: CE_POINTER [INTEGER_8]
		do
			Result := data
		end

	print2
		do
			print_raw (data, 0)
		end

	init_size (a_size: INTEGER_32)
		do
			if a_size = 0 then
				Current.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			else
				Current.data := create {CE_INTEGER_8_POINTER}.make_cast ((bson_malloc (a_size)))
			end
			Current.data_size := a_size
			Current.cur := data + 4;
			((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.reset
		end

	init
		do
			((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.init_size (initial_buffer_size.item)
		end

	init_size2 (a_size: INTEGER_32)
		do
			((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.init_size (a_size)
		end

	append_byte (a_c: INTEGER_8)
		do
			cur[0] := a_c
			Current.cur := cur + 1
		end

	append (a_data_0: CE_POINTER [ANY]; a_len: INTEGER_32)
		do
			memcpy (create {CE_VOID_POINTER}.make_cast (cur), a_data_0, a_len.to_natural_32).do_nothing
			Current.cur := cur + a_len
		end

	append32 (a_data_0: CE_POINTER [ANY])
		do
			memcpy (create {CE_VOID_POINTER}.make_cast (cur), a_data_0, (4).to_natural_32).do_nothing
			Current.cur := cur + (4)
		end

	append64 (a_data_0: CE_POINTER [ANY])
		do
			memcpy (create {CE_VOID_POINTER}.make_cast (cur), a_data_0, (8).to_natural_32).do_nothing
			Current.cur := cur + (8)
		end

	ensure_space (a_bytes_needed: INTEGER_32): INTEGER_32
		local
			l_pos2: INTEGER_32
			l_orig: CE_POINTER [INTEGER_8]
			l_new_size: INTEGER_32
			l_return: BOOLEAN
		do
			l_pos2 := cur.distance (data)
			l_orig := data
			if l_pos2 + a_bytes_needed <= data_size then
				Result := 0
				l_return := True
			end
			if not l_return then
				l_new_size := (1.5 * (data_size + a_bytes_needed).to_double).truncated_to_integer
				if l_new_size < data_size then
					if data_size + a_bytes_needed < 2_147_483_647 then
						l_new_size := 2_147_483_647
					else
						Current.err := 1
						Result := (-1)
						l_return := True
					end
				end
			end
			if not l_return then
				Current.data := create {CE_INTEGER_8_POINTER}.make_cast ((bson_realloc (create {CE_VOID_POINTER}.make_cast (data), l_new_size)))
				if data.is_void then
					bson_fatal_msg (((not data.is_void)).to_integer, ce_cstring ("realloc() failed"))
				end
				Current.data_size := l_new_size
				Current.cur := cur + (data.distance (l_orig))
				Result := 0
			end
		end

	finish: INTEGER_32
		local
			l_i: CE_ARRAY [INTEGER_32]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_32_ARRAY} l_i.make (<<1>>)

			if ((err).bit_and (2)).to_boolean then
				Result := (-1)
				l_return := True
			end
			if not l_return and then (not finished.to_boolean) then
				if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.ensure_space (1)) = -1 then
					Result := (-1)
					l_return := True
				end
				if not l_return then
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_byte ((0).to_integer_8)
					l_i.item := cur.distance (data)
					memcpy (create {CE_VOID_POINTER}.make_cast (data), create {CE_VOID_POINTER}.make_cast (l_i), (4).to_natural_32).do_nothing
					Current.finished := 1
				end
			end
			if not l_return then
				Result := 0
			end
		end

	append_estart (a_type: INTEGER_32; a_name: CE_POINTER [INTEGER_8]; a_data_size: INTEGER_32): INTEGER_32
		require
			not a_name.is_void
		local
			l_len: INTEGER_32
			l_return: BOOLEAN
		do
			l_len := ((strlen (a_name)) + {NATURAL_32} 1).to_integer_32
			if finished.to_boolean then
				Current.err := (err).bit_or ((16))
				Result := (-1)
				l_return := True
			end
			if not l_return then
				if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.ensure_space (((1 + l_len) + a_data_size))) = -1 then
					Result := (-1)
					l_return := True
				end
			end
			if not l_return then
				if (check_field_name ((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer)), a_name, (l_len - 1))) = -1 then
					builder_error ((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer)))
					Result := (-1)
					l_return := True
				end
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_byte (a_type.to_integer_8);
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_name), l_len)
				Result := 0
			end
		end

	append_int (a_name: CE_POINTER [INTEGER_8]; a_i: INTEGER_32): INTEGER_32
		require
			not a_name.is_void
		local
			l_i: CE_ARRAY [INTEGER_32]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_32_ARRAY} l_i.make (<<1>>)

			l_i.item := a_i
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (16, a_name, 4)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_i))
				Result := 0
			end
		end

	append_double (a_name: CE_POINTER [INTEGER_8]; a_d: REAL_64): INTEGER_32
		require
			not a_name.is_void
		local
			l_d: CE_ARRAY [REAL_64]
			l_return: BOOLEAN
		do
			create {CE_REAL_64_ARRAY} l_d.make (<<1>>)

			l_d.item := a_d
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (1, a_name, 8)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append64 (create {CE_VOID_POINTER}.make_cast (l_d))
				Result := 0
			end
		end

	append_bool (a_name: CE_POINTER [INTEGER_8]; a_i: INTEGER_32): INTEGER_32
		require
			not a_name.is_void
		local
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (8, a_name, 1)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_byte ((a_i /= 0).to_integer.to_integer_8)
				Result := 0
			end
		end

	append_null (a_name: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_name.is_void
		local
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (10, a_name, 0)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				Result := 0
			end
		end

	append_undefined (a_name: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_name.is_void
		local
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (6, a_name, 0)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				Result := 0
			end
		end

	append_code_w_scope_n (a_name, a_code: CE_POINTER [INTEGER_8]; a_len: INTEGER_32; a_scope: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_name.is_void
		local
			l_sl: CE_ARRAY [INTEGER_32]
			l_size: CE_ARRAY [INTEGER_32]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_32_ARRAY} l_sl.make (<<1>>)
			create {CE_INTEGER_32_ARRAY} l_size.make (<<1>>)

			l_sl.item := a_len + 1
			l_size.item := (8 + l_sl.item) + (bson_size (a_scope))
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (15, a_name, l_size.item)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_size));
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_sl));
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_code), l_sl.item);
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_scope.item.data), (bson_size (a_scope)))
				Result := 0
			end
		end

	append_code_w_scope (a_name, a_code: CE_POINTER [INTEGER_8]; a_scope: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_name.is_void
			not a_code.is_void
		do
			Result := (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_code_w_scope_n (a_name, a_code, (strlen (a_code)).to_integer_32, a_scope))
		end

	append_binary (a_name: CE_POINTER [INTEGER_8]; a_type: INTEGER_8; a_str: CE_POINTER [INTEGER_8]; a_len: INTEGER_32): INTEGER_32
		require
			not a_name.is_void
		local
			l_len: CE_ARRAY [INTEGER_32]
			l_subtwolen: CE_ARRAY [INTEGER_32]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_32_ARRAY} l_len.make (<<1>>)
			create {CE_INTEGER_32_ARRAY} l_subtwolen.make (<<1>>)

			l_len.item := a_len
			if a_type.to_integer_32 = 2 then
				l_subtwolen.item := l_len.item + 4
				if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (5, a_name, (9 + l_len.item))) = -1 then
					Result := (-1)
					l_return := True
				end
				if not l_return then
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_subtwolen));
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_byte (a_type);
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_len));
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_str), l_len.item)
				end
			else
				if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (5, a_name, (5 + l_len.item))) = -1 then
					Result := (-1)
					l_return := True
				end
				if not l_return then
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_len));
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_byte (a_type);
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_str), l_len.item)
				end
			end
			if not l_return then
				Result := 0
			end
		end

	append_oid (a_name: CE_POINTER [INTEGER_8]; a_oid: CE_POINTER [PD_BSON_OID_T]): INTEGER_32
		require
			not a_name.is_void
		local
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (7, a_name, 12)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_oid), 12)
				Result := 0
			end
		end

	append_new_oid (a_name: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_name.is_void
		local
			l_oid: CE_ARRAY [PD_BSON_OID_T]
		do
			create {CE_CLASS_ARRAY [PD_BSON_OID_T]} l_oid.make (<<1>>)

			l_oid.item.bson_oid_gen
			Result := (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_oid (a_name, l_oid))
		end

	append_regex (a_name, a_pattern, a_opts: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_name.is_void
			not a_pattern.is_void
			not a_opts.is_void
		local
			l_plen: INTEGER_32
			l_olen: INTEGER_32
			l_return: BOOLEAN
		do
			l_plen := ((strlen (a_pattern)) + {NATURAL_32} 1).to_integer_32
			l_olen := ((strlen (a_opts)) + {NATURAL_32} 1).to_integer_32
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (11, a_name, (l_plen + l_olen))) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				if ((create {P_SRC_ENCODING_STATIC}).bson_check_string ((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer)), a_pattern, (l_plen - 1))) = -1 then
					Result := (-1)
					l_return := True
				end
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_pattern), l_plen);
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_opts), l_olen)
				Result := 0
			end
		end

	append_bson (a_name: CE_POINTER [INTEGER_8]; a_bson_0: CE_POINTER [PD_BSON]): INTEGER_32
		require
			not a_name.is_void
		local
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (3, a_name, (bson_size (a_bson_0)))) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_bson_0.item.data), (bson_size (a_bson_0)))
				Result := 0
			end
		end

	append_element (a_name_or_null: CE_POINTER [INTEGER_8]; a_elem: CE_POINTER [PD_BSON_ITERATOR]): INTEGER_32
		require
			not a_elem.is_void
		local
			l_next: CE_ARRAY [PD_BSON_ITERATOR]
			l_size: INTEGER_32
			l_data_size: INTEGER_32
			l_return: BOOLEAN
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_next.make (<<1>>)

			l_next.item := (a_elem.item).twin
			l_next.item.next.do_nothing
			l_size := (l_next.item.cur).distance ((a_elem.item.cur))
			if a_name_or_null.is_void then
				if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.ensure_space (l_size)) = -1 then
					Result := (-1)
					l_return := True
				end
				if not l_return then
					((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast (a_elem.item.cur), l_size)
				end
			else
				l_data_size := ((l_size - 2).to_natural_32 - (strlen ((a_elem.item.key)))).to_integer_32;
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart ((a_elem.item.cur[0]).to_integer_32, a_name_or_null, l_data_size).do_nothing;
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append (create {CE_VOID_POINTER}.make_cast ((a_elem.item.value)), l_data_size)
			end
			if not l_return then
				Result := 0
			end
		end

	append_timestamp (a_name: CE_POINTER [INTEGER_8]; a_ts: CE_POINTER [PD_BSON_TIMESTAMP_T]): INTEGER_32
		require
			not a_name.is_void
		local
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (17, a_name, 8)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (a_ts.c_pointer + (create {PD_BSON_TIMESTAMP_T}.make).offset_of ("i")))));
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (a_ts.c_pointer + (create {PD_BSON_TIMESTAMP_T}.make).offset_of ("t")))))
				Result := 0
			end
		end

	append_timestamp2 (a_name: CE_POINTER [INTEGER_8]; a_time_0, a_increment: INTEGER_32): INTEGER_32
		require
			not a_name.is_void
		local
			l_increment: CE_ARRAY [INTEGER_32]
			l_time_0: CE_ARRAY [INTEGER_32]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_32_ARRAY} l_increment.make (<<1>>)
			create {CE_INTEGER_32_ARRAY} l_time_0.make (<<1>>)

			l_increment.item := a_increment
			l_time_0.item := a_time_0
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (17, a_name, 8)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_increment));
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (l_time_0))
				Result := 0
			end
		end

	append_date (a_name: CE_POINTER [INTEGER_8]; a_millis: INTEGER_64): INTEGER_32
		require
			not a_name.is_void
		local
			l_millis: CE_ARRAY [INTEGER_64]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_64_ARRAY} l_millis.make (<<1>>)

			l_millis.item := a_millis
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (9, a_name, 8)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append64 (create {CE_VOID_POINTER}.make_cast (l_millis))
				Result := 0
			end
		end

	append_time_t (a_name: CE_POINTER [INTEGER_8]; a_secs: INTEGER_32): INTEGER_32
		require
			not a_name.is_void
		do
			Result := (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_date (a_name, a_secs.to_integer_64 * {INTEGER_64} 1_000))
		end

	append_start_object (a_name: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_name.is_void
		local
			l_tmp_0: INTEGER_32
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (3, a_name, 5)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				l_tmp_0 := stack_pos
				Current.stack_pos := stack_pos + 1
				stack[l_tmp_0] := cur.distance (data);
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (zero.item))
				Result := 0
			end
		end

	append_start_array (a_name: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_name.is_void
		local
			l_tmp_0: INTEGER_32
			l_return: BOOLEAN
		do
			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_estart (4, a_name, 5)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				l_tmp_0 := stack_pos
				Current.stack_pos := stack_pos + 1
				stack[l_tmp_0] := cur.distance (data);
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append32 (create {CE_VOID_POINTER}.make_cast (zero.item))
				Result := 0
			end
		end

	append_finish_object: INTEGER_32
		local
			l_start: CE_POINTER [INTEGER_8]
			l_i: CE_ARRAY [INTEGER_32]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_32_ARRAY} l_i.make (<<1>>)

			if (((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.ensure_space (1)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_byte ((0).to_integer_8)
				Current.stack_pos := stack_pos - 1
				l_start := data + stack[stack_pos]
				l_i.item := cur.distance (l_start)
				memcpy (create {CE_VOID_POINTER}.make_cast (l_start), create {CE_VOID_POINTER}.make_cast (l_i), (4).to_natural_32).do_nothing
				Result := 0
			end
		end

	append_finish_array: INTEGER_32
		do
			Result := ((create {CE_CLASS_POINTER [PD_BSON]}.make_shared (c_pointer))).item.append_finish_object
		end

	append_string (a_name, a_value: CE_POINTER [INTEGER_8]): INTEGER_32
		require
			not a_value.is_void
		do
			Result := bson_append_string_base (address, a_name, a_value, (strlen (a_value)).to_integer_32, {PC_BSON_TYPE}.Bson_string)
		end

	destroy
		do
			bson_free (create {CE_VOID_POINTER}.make_cast (data))
			Current.err := 0
			Current.data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			Current.cur := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			Current.finished := 1
		end

feature {PD_BSON} -- private routines

	print_raw (a_data_0: CE_POINTER [INTEGER_8]; a_depth: INTEGER_32)
		local
			l_i: CE_ARRAY [PD_BSON_ITERATOR]
			l_key: CE_POINTER [INTEGER_8]
			l_temp: INTEGER_32
			l_ts: PD_BSON_TIMESTAMP_T
			l_oidhex: CE_ARRAY [INTEGER_8]
			l_scope: CE_ARRAY [PD_BSON]
			l_t: NATURAL_32
			l_tmp_5: CE_POINTER [INTEGER_8]
			l_pos: INTEGER_32
			l_return, l_break: BOOLEAN
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_i.make (<<1>>)
			create {CE_INTEGER_8_ARRAY} l_oidhex.make (<<25>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_scope.make (<<1>>)

			from until l_pos = -1 loop -- -1: return
				inspect l_pos

				-- <function body>
				when 0 then -- execution start:
					l_i.item.from_buffer (a_data_0)
				when 1 then -- loop condition:
					if (l_return or l_break) then
						l_pos := -1 + 23 -- skip loop
					end
				when 2 then -- loop body:
						l_break := False
						if not (l_i.item.next).to_boolean then
							l_break := True
						end
						if not (l_return or l_break) then
							l_t := (l_i.item.type)
							if l_t = {NATURAL_32} 0 then
								l_break := True
							end
						end
						if (l_return or l_break) then
							l_pos := -1 + 22 -- skip if block
						end
				when 3 then -- if:
						l_key := l_i.item.key
						l_temp := 0
						from
						until
							not (l_temp <= a_depth)
						loop
							printf2.item.call ([ce_cstring ("%T")])
							l_temp := l_temp + 1
						end
						printf2.item.call ([ce_cstring ("%%s : %%d %T "), l_key, l_t])
						inspect l_t -- switch 1 head
						when 1 then l_pos := -1 + 4
						when 2 then l_pos := -1 + 5
						when 14 then l_pos := -1 + 6
						when 7 then l_pos := -1 + 7
						when 8 then l_pos := -1 + 8
						when 9 then l_pos := -1 + 9
						when 5 then l_pos := -1 + 10
						when 6 then l_pos := -1 + 11
						when 10 then l_pos := -1 + 12
						when 11 then l_pos := -1 + 13
						when 13 then l_pos := -1 + 14
						when 15 then l_pos := -1 + 15
						when 16 then l_pos := -1 + 16
						when 18 then l_pos := -1 + 17
						when 17 then l_pos := -1 + 18
						when 4 then l_pos := -1 + 19
						when 3 then l_pos := -1 + 19
						else l_pos := -1 + 20
						end
				when 4 then -- switch 1, case 1u
						printf2.item.call ([ce_cstring ("%%f"), (l_i.item.double)])
						l_pos := -1 + 21 -- switch 1; break
				when 5 then -- switch 1, case 2u
						printf2.item.call ([ce_cstring ("%%s"), (l_i.item.string)])
						l_pos := -1 + 21 -- switch 1; break
				when 6 then -- switch 1, case 14u
						printf2.item.call ([ce_cstring ("SYMBOL: %%s"), (l_i.item.string)])
						l_pos := -1 + 21 -- switch 1; break
				when 7 then -- switch 1, case 7u
						oid_to_string ((l_i.item.oid), l_oidhex)
						printf2.item.call ([ce_cstring ("%%s"), l_oidhex])
						l_pos := -1 + 21 -- switch 1; break
				when 8 then -- switch 1, case 8u
						if (l_i.item.bool).to_boolean then
							l_tmp_5 := ce_cstring ("true")
						else
							l_tmp_5 := ce_cstring ("false")
						end
						printf2.item.call ([ce_cstring ("%%s"), l_tmp_5])
						l_pos := -1 + 21 -- switch 1; break
				when 9 then -- switch 1, case 9u
						printf2.item.call ([ce_cstring ("%%ld"), (l_i.item.date).to_integer_32])
						l_pos := -1 + 21 -- switch 1; break
				when 10 then -- switch 1, case 5u
						printf2.item.call ([ce_cstring ("BSON_BINDATA")])
						l_pos := -1 + 21 -- switch 1; break
				when 11 then -- switch 1, case 6u
						printf2.item.call ([ce_cstring ("BSON_UNDEFINED")])
						l_pos := -1 + 21 -- switch 1; break
				when 12 then -- switch 1, case 10u
						printf2.item.call ([ce_cstring ("BSON_NULL")])
						l_pos := -1 + 21 -- switch 1; break
				when 13 then -- switch 1, case 11u
						printf2.item.call ([ce_cstring ("BSON_REGEX: %%s"), (iterator_regex (l_i))])
						l_pos := -1 + 21 -- switch 1; break
				when 14 then -- switch 1, case 13u
						printf2.item.call ([ce_cstring ("BSON_CODE: %%s"), (iterator_code (l_i))])
						l_pos := -1 + 21 -- switch 1; break
				when 15 then -- switch 1, case 15u
						printf2.item.call ([ce_cstring ("BSON_CODE_W_SCOPE: %%s"), (iterator_code (l_i))])
						l_scope.item.init
						iterator_code_scope (l_i, l_scope)
						printf2.item.call ([ce_cstring ("%N%T SCOPE: ")])
						l_scope.item.print2
						l_pos := -1 + 21 -- switch 1; break
				when 16 then -- switch 1, case 16u
						printf2.item.call ([ce_cstring ("%%d"), (l_i.item.int)])
						l_pos := -1 + 21 -- switch 1; break
				when 17 then -- switch 1, case 18u
						printf2.item.call ([ce_cstring ("%%lld"), (l_i.item.long).to_natural_64])
						l_pos := -1 + 21 -- switch 1; break
				when 18 then -- switch 1, case 17u
						l_ts := (iterator_timestamp (l_i)).twin
						printf2.item.call ([ce_cstring ("i: %%d, t: %%d"), l_ts.i, l_ts.t])
						l_pos := -1 + 21 -- switch 1; break
				when 19 then -- switch 1, case 4u | switch 1, case 3u
						printf2.item.call ([ce_cstring ("%N")])
						print_raw ((l_i.item.value), a_depth + 1)
						l_pos := -1 + 21 -- switch 1; break
				when 20 then -- switch 1, default:;
						bson_errprintf.item.call ([ce_cstring ("can%'t print type : %%d%N"), l_t])
				when 21 then -- switch 1, done:
						printf2.item.call ([ce_cstring ("%N")])
				when 22 then -- if done
						l_pos := -1 + 1 -- back to loop condition
				when 23 then -- loop done
					l_break := False
				when 24 then
					l_pos := -1 + -1 -- end of function
				end
				-- </function body>

				l_pos := l_pos + 1
			end
		end

	oid_to_string (a_oid: CE_POINTER [PD_BSON_OID_T]; a_str: CE_POINTER [INTEGER_8])
		require
			not a_str.is_void
		local
			l_i: INTEGER_32
		do
			l_i := 0
			from
			until
				not (l_i < 12)
			loop
				a_str[2 * l_i] := hex.item[((((a_oid.item.bytes[l_i]).to_integer_32).bit_and (240))).bit_shift_right (4)]
				a_str[(2 * l_i + 1)] := hex.item[((a_oid.item.bytes[l_i]).to_integer_32).bit_and (15)]
				l_i := l_i + 1
			end
			a_str[24] := (('%/0/').code).to_integer_8
		end

	iterator_timestamp (a_i: CE_POINTER [PD_BSON_ITERATOR]): PD_BSON_TIMESTAMP_T
		require
			not a_i.is_void
		local
			l_ts: CE_ARRAY [PD_BSON_TIMESTAMP_T]
		do
			create {CE_CLASS_ARRAY [PD_BSON_TIMESTAMP_T]} l_ts.make (<<1>>)

			memcpy (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (l_ts.item.c_pointer + (create {PD_BSON_TIMESTAMP_T}.make).offset_of ("i")))), create {CE_VOID_POINTER}.make_cast ((a_i.item.value)), (4).to_natural_32).do_nothing
			memcpy (create {CE_VOID_POINTER}.make_cast ((create {CE_INTEGER_32_POINTER}.make_shared (l_ts.item.c_pointer + (create {PD_BSON_TIMESTAMP_T}.make).offset_of ("t")))), create {CE_VOID_POINTER}.make_cast (((a_i.item.value) + 4)), (4).to_natural_32).do_nothing
			Result := (l_ts.item).twin
		end

	iterator_code (a_i: CE_POINTER [PD_BSON_ITERATOR]): CE_POINTER [INTEGER_8]
		local
			l_pos: INTEGER_32
			l_return: BOOLEAN
		do
			create {CE_INTEGER_8_POINTER} Result.make

			from until l_pos = -1 loop -- -1: return
				inspect l_pos

				-- <function body>
				when 0 then -- execution start:
					inspect (a_i.item.type) -- switch 1 head
					when 13 then l_pos := -1 + 1
					when 2 then l_pos := -1 + 1
					when 15 then l_pos := -1 + 2
					else l_pos := -1 + 3
					end
				when 1 then -- switch 1, case 13u | switch 1, case 2u
					Result := ((a_i.item.value) + 4)
					l_return := True
				when 2 then -- switch 1, case 15u
					if not l_return then
						Result := ((a_i.item.value) + 8)
						l_return := True
					end
				when 3 then -- switch 1, default:
					if not l_return then
						Result := (create {CE_INTEGER_8_POINTER}.make_shared (default_pointer))
					end
				when 4 then -- switch 1, done:
					l_pos := -1 + -1 -- end of function
				end
				-- </function body>

				l_pos := l_pos + 1
			end
		end

	iterator_code_scope (a_i: CE_POINTER [PD_BSON_ITERATOR]; a_scope: CE_POINTER [PD_BSON])
		require
			not a_i.is_void
			not a_scope.is_void
		local
			l_code_len: CE_ARRAY [INTEGER_32]
		do
			create {CE_INTEGER_32_ARRAY} l_code_len.make (<<1>>)

			if (a_i.item.type) = {NATURAL_32} 15 then
				memcpy (create {CE_VOID_POINTER}.make_cast (l_code_len), create {CE_VOID_POINTER}.make_cast (((a_i.item.value) + 4)), (4).to_natural_32).do_nothing
				a_scope.item.init_data ((((a_i.item.value) + 8) + l_code_len.item)).do_nothing
				a_scope.item.reset
				a_scope.item.finished := 1
			else
				a_scope.item.empty.do_nothing
			end
		end

	iterator_regex (a_i: CE_POINTER [PD_BSON_ITERATOR]): CE_POINTER [INTEGER_8]
		require
			not a_i.is_void
		do
			Result := (a_i.item.value)
		end

	builder_error (a_b: CE_POINTER [PD_BSON])
		do
			if (not (err_handler.item).is_void) then
				err_handler.item.call ([ce_cstring ("BSON error.")])
			end
		end

	check_field_name (a_b: CE_POINTER [PD_BSON]; a_string: CE_POINTER [INTEGER_8]; a_length: INTEGER_32): INTEGER_32
		do
			Result := ((create {P_SRC_ENCODING_STATIC}).bson_validate_string (a_b, create {CE_NATURAL_8_POINTER}.make_cast (a_string), a_length, (1).to_integer_8, (1).to_integer_8, (1).to_integer_8))
		end

feature {ANY}

	structure_size: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return sizeof(struct __anonstruct_bson_11);"
		end

	structure_alignment: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return __alignof__(struct __anonstruct_bson_11);"
		end

	offset_of (a_field_name: STRING_8): INTEGER_32
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_field_name)
			Result := c__offset_of (l_c_str.item)
		end

feature {ANY}

	copy (a_other: like Current)
		do
			make
			internal_c_pointer.memory_copy (a_other.c_pointer, structure_size)
		end

	is_equal (a_other: like Current): BOOLEAN
		do
			Result := internal_c_pointer.memory_compare (a_other.c_pointer, structure_size)
		end

feature {NONE}

	structure_info: CE_STRUCTURE
		do
			-- dummy for CE_CLASS invariant
			create Result.make_struct (<<>>)
			Result.alignment := structure_alignment
		end

	c__data (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_bson_11 *)$a_c_pointer)->data;"
		end

	c__cur (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_bson_11 *)$a_c_pointer)->cur;"
		end

	c__data_size (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_bson_11 *)$a_c_pointer)->dataSize;"
		end

	c__finished (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_bson_11 *)$a_c_pointer)->finished;"
		end

	c__stack (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_bson_11 *)$a_c_pointer)->stack;"
		end

	c__stack_pos (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_bson_11 *)$a_c_pointer)->stackPos;"
		end

	c__err (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_bson_11 *)$a_c_pointer)->err;"
		end

	c__errstr (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_bson_11 *)$a_c_pointer)->errstr;"
		end

	c__set_data (a_c_pointer: POINTER; a_data: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_bson_11 *)$a_c_pointer)->data = $a_data;"
		end

	c__set_cur (a_c_pointer: POINTER; a_cur: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_bson_11 *)$a_c_pointer)->cur = $a_cur;"
		end

	c__set_data_size (a_c_pointer: POINTER; a_data_size: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_bson_11 *)$a_c_pointer)->dataSize = $a_data_size;"
		end

	c__set_finished (a_c_pointer: POINTER; a_finished: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_bson_11 *)$a_c_pointer)->finished = $a_finished;"
		end

	c__set_stack_pos (a_c_pointer: POINTER; a_stack_pos: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_bson_11 *)$a_c_pointer)->stackPos = $a_stack_pos;"
		end

	c__set_err (a_c_pointer: POINTER; a_err: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_bson_11 *)$a_c_pointer)->err = $a_err;"
		end

	c__set_errstr (a_c_pointer: POINTER; a_errstr: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_bson_11 *)$a_c_pointer)->errstr = $a_errstr;"
		end

	c__offset_of (a_field_name: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%", <string.h>, <stddef.h>, <stdio.h>"
		alias
			"[
				if(strcmp($a_field_name, "data") == 0) return offsetof(struct __anonstruct_bson_11, data);
				if(strcmp($a_field_name, "cur") == 0) return offsetof(struct __anonstruct_bson_11, cur);
				if(strcmp($a_field_name, "data_size") == 0) return offsetof(struct __anonstruct_bson_11, dataSize);
				if(strcmp($a_field_name, "finished") == 0) return offsetof(struct __anonstruct_bson_11, finished);
				if(strcmp($a_field_name, "stack") == 0) return offsetof(struct __anonstruct_bson_11, stack);
				if(strcmp($a_field_name, "stack_pos") == 0) return offsetof(struct __anonstruct_bson_11, stackPos);
				if(strcmp($a_field_name, "err") == 0) return offsetof(struct __anonstruct_bson_11, err);
				if(strcmp($a_field_name, "errstr") == 0) return offsetof(struct __anonstruct_bson_11, errstr);
				fprintf(stderr, "struct __anonstruct_bson_11.%s not found\n", (char *)$a_field_name);
				return -1;
			]"
		end

end
