class
	PD_GRIDFILE

inherit
	CE_UTILITIES
		redefine
			copy,
			is_equal
		end
	CE_CLASS
		redefine
			make,
			make_shared,
			structure_size,
			offset_of,
			copy,
			is_equal
		end
	S_STRING_STATIC
		redefine
			copy,
			is_equal
		end
	S_STDIO_STATIC
		redefine
			copy,
			is_equal
		end

create
	make,
	make_shared,
	make_initialized

feature {NONE}

	make
		do
			internal_c_pointer := internal_c_pointer.memory_alloc (structure_size)
			is_allocated_memory := True
		end

	make_shared (a_c_pointer: POINTER)
		do
			internal_c_pointer := a_c_pointer
		end

	make_initialized (a_data: TUPLE [
				a_gfs: CE_POINTER [PD_GRIDFS]
				a_meta: CE_POINTER [PD_BSON]
				a_pos: NATURAL_64
				a_id: PD_BSON_OID_T
				a_remote_name: CE_POINTER [INTEGER_8]
				a_content_type: CE_POINTER [INTEGER_8]
				a_length: NATURAL_64
				a_chunk_num: INTEGER_32
				a_pending_data: CE_POINTER [INTEGER_8]
				a_pending_len: INTEGER_32
			])
		require else
			a_data /= Void
		do
			make
			Current.gfs := a_data.a_gfs
			Current.meta := a_data.a_meta
			Current.pos := a_data.a_pos
			id.c_pointer.memory_copy (a_data.a_id.c_pointer, a_data.a_id.structure_size)
			Current.remote_name := a_data.a_remote_name
			Current.content_type := a_data.a_content_type
			Current.length := a_data.a_length
			Current.chunk_num := a_data.a_chunk_num
			Current.pending_data := a_data.a_pending_data
			Current.pending_len := a_data.a_pending_len
		end

feature {ANY} -- fields

	gfs: CE_POINTER [PD_GRIDFS] assign set__gfs
		do
			create {CE_CLASS_POINTER [PD_GRIDFS]} Result.make_shared (c__gfs (internal_c_pointer))
		end

	meta: CE_POINTER [PD_BSON] assign set__meta
		do
			create {CE_CLASS_POINTER [PD_BSON]} Result.make_shared (c__meta (internal_c_pointer))
		end

	pos: NATURAL_64 assign set__pos
		do
			Result := c__pos (internal_c_pointer)
		end

	id: PD_BSON_OID_T assign set__id
		once ("OBJECT")
			create Result.make_shared (c__id (internal_c_pointer))
		end

	remote_name: CE_POINTER [INTEGER_8] assign set__remote_name
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__remote_name (internal_c_pointer))
		end

	content_type: CE_POINTER [INTEGER_8] assign set__content_type
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__content_type (internal_c_pointer))
		end

	length: NATURAL_64 assign set__length
		do
			Result := c__length (internal_c_pointer)
		end

	chunk_num: INTEGER_32 assign set__chunk_num
		do
			Result := c__chunk_num (internal_c_pointer)
		end

	pending_data: CE_POINTER [INTEGER_8] assign set__pending_data
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__pending_data (internal_c_pointer))
		end

	pending_len: INTEGER_32 assign set__pending_len
		do
			Result := c__pending_len (internal_c_pointer)
		end

feature {ANY} -- field setters

	set__gfs (a_gfs: CE_POINTER [PD_GRIDFS])
		do
			c__set_gfs (internal_c_pointer, a_gfs.c_pointer)
		end

	set__meta (a_meta: CE_POINTER [PD_BSON])
		do
			c__set_meta (internal_c_pointer, a_meta.c_pointer)
		end

	set__pos (a_pos: NATURAL_64)
		do
			c__set_pos (internal_c_pointer, a_pos)
		end

	set__id (a_id: PD_BSON_OID_T)
		do
			c__set_id (internal_c_pointer, a_id.c_pointer)
		end

	set__remote_name (a_remote_name: CE_POINTER [INTEGER_8])
		do
			c__set_remote_name (internal_c_pointer, a_remote_name.c_pointer)
		end

	set__content_type (a_content_type: CE_POINTER [INTEGER_8])
		do
			c__set_content_type (internal_c_pointer, a_content_type.c_pointer)
		end

	set__length (a_length: NATURAL_64)
		do
			c__set_length (internal_c_pointer, a_length)
		end

	set__chunk_num (a_chunk_num: INTEGER_32)
		do
			c__set_chunk_num (internal_c_pointer, a_chunk_num)
		end

	set__pending_data (a_pending_data: CE_POINTER [INTEGER_8])
		do
			c__set_pending_data (internal_c_pointer, a_pending_data.c_pointer)
		end

	set__pending_len (a_pending_len: INTEGER_32)
		do
			c__set_pending_len (internal_c_pointer, a_pending_len)
		end

feature {ANY} -- public routines

	get_descriptor (a_out: CE_POINTER [PD_BSON])
		require
			not a_out.is_void
		do
			a_out.item := (meta.item).twin
		end

	writer_init (a_gfs: CE_POINTER [PD_GRIDFS]; a_remote_name, a_content_type: CE_POINTER [INTEGER_8])
		require
			not a_remote_name.is_void
			not a_content_type.is_void
		do
			Current.gfs := a_gfs;
			(create {CE_CLASS_POINTER [PD_BSON_OID_T]}.make_shared (c_pointer + (create {PD_GRIDFILE}.make).offset_of ("id"))).item.bson_oid_gen
			Current.chunk_num := 0
			Current.length := (0).to_natural_64
			Current.pending_len := 0
			Current.pending_data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
			Current.remote_name := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (((strlen (a_remote_name)) + {NATURAL_32} 1).to_integer_32)))
			strcpy (remote_name, a_remote_name).do_nothing
			Current.content_type := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (((strlen (a_content_type)) + {NATURAL_32} 1).to_integer_32)))
			strcpy (content_type, a_content_type).do_nothing
		end

	write_buffer (a_data_0: CE_POINTER [INTEGER_8]; a_length: NATURAL_64)
		local
			l_data_0: CE_POINTER [INTEGER_8]
			l_bytes_left: INTEGER_32
			l_data_partial_len: INTEGER_32
			l_chunks_to_write: INTEGER_32
			l_buffer: CE_POINTER [INTEGER_8]
			l_o_chunk: CE_POINTER [PD_BSON]
			l_to_write: NATURAL_64
		do
			l_data_0 := a_data_0
			l_bytes_left := 0
			l_data_partial_len := 0
			l_chunks_to_write := 0
			l_to_write := a_length + pending_len.to_natural_64
			if l_to_write < {NATURAL_64} 262_144 then
				if (not pending_data.is_void) then
					Current.pending_data := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_realloc (create {CE_VOID_POINTER}.make_cast (pending_data), (pending_len.to_natural_64 + l_to_write).to_integer_32)))
					memcpy (create {CE_VOID_POINTER}.make_cast ((pending_data + pending_len)), create {CE_VOID_POINTER}.make_cast (l_data_0), a_length.to_natural_32).do_nothing
				elseif l_to_write > {NATURAL_64} 0 then
					Current.pending_data := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (l_to_write.to_integer_32)))
					memcpy (create {CE_VOID_POINTER}.make_cast (pending_data), create {CE_VOID_POINTER}.make_cast (l_data_0), a_length.to_natural_32).do_nothing
				end
				Current.pending_len := (pending_len.to_natural_64 + a_length).to_integer_32
			else
				l_chunks_to_write := (l_to_write // {NATURAL_64} 262_144).to_integer_32
				l_bytes_left := (l_to_write \\ {NATURAL_64} 262_144).to_integer_32
				if pending_len > 0 then
					l_data_partial_len := 262_144 - pending_len
					l_buffer := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (262_144)))
					memcpy (create {CE_VOID_POINTER}.make_cast (l_buffer), create {CE_VOID_POINTER}.make_cast (pending_data), pending_len.to_natural_32).do_nothing
					memcpy (create {CE_VOID_POINTER}.make_cast ((l_buffer + pending_len)), create {CE_VOID_POINTER}.make_cast (l_data_0), l_data_partial_len.to_natural_32).do_nothing
					l_o_chunk := (create {P_SRC_GRIDFS_STATIC}).chunk_new (id.twin, chunk_num, l_buffer, 262_144);
					gfs.item.client.item.insert (gfs.item.chunks_ns, l_o_chunk, create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer)).do_nothing;
					(create {P_SRC_GRIDFS_STATIC}).chunk_free (l_o_chunk)
					Current.chunk_num := chunk_num + 1
					Current.length := length + ({NATURAL_64} 262_144)
					l_data_0 := l_data_0 + l_data_partial_len
					l_chunks_to_write := l_chunks_to_write - 1;
					(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (l_buffer))
				end
				from
				until
					not (l_chunks_to_write > 0)
				loop
					l_o_chunk := (create {P_SRC_GRIDFS_STATIC}).chunk_new (id.twin, chunk_num, l_data_0, 262_144);
					gfs.item.client.item.insert (gfs.item.chunks_ns, l_o_chunk, create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer)).do_nothing;
					(create {P_SRC_GRIDFS_STATIC}).chunk_free (l_o_chunk)
					Current.chunk_num := chunk_num + 1
					l_chunks_to_write := l_chunks_to_write - 1
					Current.length := length + ({NATURAL_64} 262_144)
					l_data_0 := l_data_0 + (262_144)
				end
				(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (pending_data))
				if l_bytes_left = 0 then
					Current.pending_data := create {CE_INTEGER_8_POINTER}.make_shared (default_pointer)
				else
					Current.pending_data := create {CE_INTEGER_8_POINTER}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (l_bytes_left)))
					memcpy (create {CE_VOID_POINTER}.make_cast (pending_data), create {CE_VOID_POINTER}.make_cast (l_data_0), l_bytes_left.to_natural_32).do_nothing
				end
				Current.pending_len := l_bytes_left
			end
		end

	writer_done: INTEGER_32
		local
			l_o_chunk: CE_POINTER [PD_BSON]
			l_response: INTEGER_32
		do
			if (not pending_data.is_void) then
				l_o_chunk := (create {P_SRC_GRIDFS_STATIC}).chunk_new (id.twin, chunk_num, pending_data, pending_len);
				gfs.item.client.item.insert (gfs.item.chunks_ns, l_o_chunk, create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer)).do_nothing;
				(create {P_SRC_GRIDFS_STATIC}).chunk_free (l_o_chunk);
				(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (pending_data))
				Current.length := length + pending_len.to_natural_64
			end
			l_response := gfs.item.insert_file (remote_name, id.twin, length, content_type);
			(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (remote_name));
			(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (content_type))
			Result := l_response
		end

	destroy
		do
			meta.item.destroy;
			(create {P_SRC_BSON_STATIC}).bson_free (create {CE_VOID_POINTER}.make_cast (meta))
		end

	get_filename: CE_POINTER [INTEGER_8]
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, ce_cstring ("filename")).do_nothing
			Result := (l_it.item.string)
		end

	get_chunksize: INTEGER_32
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, ce_cstring ("chunkSize")).do_nothing
			Result := (l_it.item.int)
		end

	get_contentlength: NATURAL_64
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, ce_cstring ("length")).do_nothing
			if (l_it.item.type) = {NATURAL_32} 16 then
				Result := (l_it.item.int).to_natural_64
			else
				Result := (l_it.item.long).to_natural_64
			end
		end

	get_contenttype: CE_POINTER [INTEGER_8]
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			if (l_it.item.bson_find (meta, ce_cstring ("contentType"))).to_boolean then
				Result := (l_it.item.string)
			else
				Result := (create {CE_INTEGER_8_POINTER}.make_shared (default_pointer))
			end
		end

	get_uploaddate: INTEGER_64
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, ce_cstring ("uploadDate")).do_nothing
			Result := (l_it.item.date)
		end

	get_md5: CE_POINTER [INTEGER_8]
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, ce_cstring ("md5")).do_nothing
			Result := (l_it.item.string)
		end

	get_field (a_name: CE_POINTER [INTEGER_8]): CE_POINTER [INTEGER_8]
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, a_name).do_nothing
			Result := (l_it.item.value)
		end

	get_boolean (a_name: CE_POINTER [INTEGER_8]): INTEGER_32
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, a_name).do_nothing
			Result := (l_it.item.bool)
		end

	get_metadata (a_out: CE_POINTER [PD_BSON])
		require
			not a_out.is_void
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			if (l_it.item.bson_find (meta, ce_cstring ("metadata"))).to_boolean then
				bson_iterator_subobject (l_it, a_out)
			else
				a_out.item.empty.do_nothing
			end
		end

	get_numchunks: INTEGER_32
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_length: NATURAL_64
			l_chunk_size: NATURAL_64
			l_numchunks: REAL_64
			l_tmp_3: INTEGER_32
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_it.item.bson_find (meta, ce_cstring ("length")).do_nothing
			if (l_it.item.type) = {NATURAL_32} 16 then
				l_length := (l_it.item.int).to_natural_64
			else
				l_length := (l_it.item.long).to_natural_64
			end
			l_it.item.bson_find (meta, ce_cstring ("chunkSize")).do_nothing
			l_chunk_size := (l_it.item.int).to_natural_64
			l_numchunks := l_length.to_real_64 / l_chunk_size.to_real_64
			if l_numchunks - (l_numchunks.truncated_to_integer).to_double > (0).to_double then
				l_tmp_3 := (l_numchunks + (1).to_double).truncated_to_integer
			else
				l_tmp_3 := l_numchunks.truncated_to_integer
			end
			Result := l_tmp_3
		end

	get_chunk (a_n: INTEGER_32; a_out: CE_POINTER [PD_BSON])
		local
			l_query: CE_ARRAY [PD_BSON]
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_id: CE_ARRAY [PD_BSON_OID_T]
			l_result: INTEGER_32
			l_empty: CE_ARRAY [PD_BSON]
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_query.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_OID_T]} l_id.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_empty.make (<<1>>)

			l_query.item.init
			l_it.item.bson_find (meta, ce_cstring ("_id")).do_nothing
			l_id.item := ((l_it.item.oid).item).twin
			l_query.item.append_oid (ce_cstring ("files_id"), l_id).do_nothing
			l_query.item.append_int (ce_cstring ("n"), a_n).do_nothing
			l_query.item.finish.do_nothing
			l_result := (((create {P_SRC_MONGO_STATIC}).mongo_find_one (gfs.item.client, gfs.item.chunks_ns, l_query, create {CE_CLASS_POINTER [PD_BSON]}.make_shared (default_pointer), a_out)) = 0).to_integer;
			l_query.item.destroy
			if not l_result.to_boolean then
				l_empty.item.empty.do_nothing;
				(create {P_SRC_BSON_STATIC}).bson_copy (a_out, l_empty).do_nothing
			end
		end

	get_chunks (a_start, a_size: INTEGER_32): CE_POINTER [PD_MONGO_CURSOR]
		local
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_id: CE_ARRAY [PD_BSON_OID_T]
			l_gte: CE_ARRAY [PD_BSON]
			l_query: CE_ARRAY [PD_BSON]
			l_orderby: CE_ARRAY [PD_BSON]
			l_command: CE_ARRAY [PD_BSON]
			l_cursor: CE_POINTER [PD_MONGO_CURSOR]
		do
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_OID_T]} l_id.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_gte.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_query.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_orderby.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_command.make (<<1>>)

			l_it.item.bson_find (meta, ce_cstring ("_id")).do_nothing
			l_id.item := ((l_it.item.oid).item).twin
			l_query.item.init
			l_query.item.append_oid (ce_cstring ("files_id"), l_id).do_nothing
			if a_size = 1 then
				l_query.item.append_int (ce_cstring ("n"), a_start).do_nothing
			else
				l_gte.item.init
				l_gte.item.append_int (ce_cstring ("$gte"), a_start).do_nothing
				l_gte.item.finish.do_nothing
				l_query.item.append_bson (ce_cstring ("n"), l_gte).do_nothing;
				l_gte.item.destroy
			end
			l_query.item.finish.do_nothing
			l_orderby.item.init
			l_orderby.item.append_int (ce_cstring ("n"), 1).do_nothing
			l_orderby.item.finish.do_nothing
			l_command.item.init
			l_command.item.append_bson (ce_cstring ("query"), l_query).do_nothing
			l_command.item.append_bson (ce_cstring ("orderby"), l_orderby).do_nothing
			l_command.item.finish.do_nothing
			l_cursor := (create {P_SRC_MONGO_STATIC}).mongo_find (gfs.item.client, gfs.item.chunks_ns, l_command, create {CE_CLASS_POINTER [PD_BSON]}.make_shared (default_pointer), a_size, 0, 0);
			l_command.item.destroy
			l_query.item.destroy
			l_orderby.item.destroy
			Result := l_cursor
		end

	write_file (a_stream: CE_POINTER [SD_IO_FILE]): NATURAL_64
		local
			l_i: INTEGER_32
			l_len: NATURAL_32
			l_chunk: CE_ARRAY [PD_BSON]
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_data_0: CE_POINTER [INTEGER_8]
			l_num: INTEGER_32
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_chunk.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_num := (((create {CE_CLASS_POINTER [PD_GRIDFILE]}.make_shared (c_pointer))).item.get_numchunks)
			l_i := 0
			from
			until
				not (l_i < l_num)
			loop
				((create {CE_CLASS_POINTER [PD_GRIDFILE]}.make_shared (c_pointer))).item.get_chunk (l_i, l_chunk)
				l_it.item.bson_find (l_chunk, ce_cstring ("data")).do_nothing
				l_len := (bson_iterator_bin_len (l_it)).to_natural_32
				l_data_0 := bson_iterator_bin_data (l_it)
				fwrite (create {CE_VOID_POINTER}.make_cast (l_data_0), {PLATFORM}.Integer_8_bytes.to_natural_32, l_len, a_stream).do_nothing;
				l_chunk.item.destroy
				l_i := l_i + 1
			end
			Result := (((create {CE_CLASS_POINTER [PD_GRIDFILE]}.make_shared (c_pointer))).item.get_contentlength)
		end

	read (a_size: NATURAL_64; a_buf: CE_POINTER [INTEGER_8]): NATURAL_64
		local
			l_size: NATURAL_64
			l_buf: CE_POINTER [INTEGER_8]
			l_chunks: CE_POINTER [PD_MONGO_CURSOR]
			l_chunk: CE_ARRAY [PD_BSON]
			l_first_chunk: INTEGER_32
			l_last_chunk: INTEGER_32
			l_total_chunks: INTEGER_32
			l_chunksize: NATURAL_64
			l_contentlength: NATURAL_64
			l_bytes_left: NATURAL_64
			l_i: INTEGER_32
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_chunk_len: NATURAL_64
			l_chunk_data: CE_POINTER [INTEGER_8]
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_chunk.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_size := a_size
			l_buf := a_buf
			l_contentlength := ((create {CE_CLASS_POINTER [PD_GRIDFILE]}.make_shared (c_pointer))).item.get_contentlength
			l_chunksize := (((create {CE_CLASS_POINTER [PD_GRIDFILE]}.make_shared (c_pointer))).item.get_chunksize).to_natural_64
			if l_contentlength - pos < l_size then
				l_size := l_contentlength - pos
			else
				l_size := l_size
			end
			l_bytes_left := l_size
			l_first_chunk := (pos // l_chunksize).to_integer_32
			l_last_chunk := (((pos + l_size) - {NATURAL_64} 1) // l_chunksize).to_integer_32
			l_total_chunks := (l_last_chunk - l_first_chunk) + 1
			l_chunks := ((create {CE_CLASS_POINTER [PD_GRIDFILE]}.make_shared (c_pointer))).item.get_chunks (l_first_chunk, l_total_chunks)
			l_i := 0
			from
			until
				not (l_i < l_total_chunks)
			loop
				l_chunks.item.next.do_nothing
				l_chunk.item := (l_chunks.item.current2).twin
				l_it.item.bson_find (l_chunk, ce_cstring ("data")).do_nothing
				l_chunk_len := (bson_iterator_bin_len (l_it)).to_natural_64
				l_chunk_data := bson_iterator_bin_data (l_it)
				if l_i = 0 then
					l_chunk_data := l_chunk_data + (pos \\ l_chunksize).to_integer_32
					l_chunk_len := l_chunk_len - (pos \\ l_chunksize)
				end
				if l_bytes_left > l_chunk_len then
					memcpy (create {CE_VOID_POINTER}.make_cast (l_buf), create {CE_VOID_POINTER}.make_cast (l_chunk_data), l_chunk_len.to_natural_32).do_nothing
					l_bytes_left := l_bytes_left - l_chunk_len
					l_buf := l_buf + l_chunk_len.to_integer_32
				else
					memcpy (create {CE_VOID_POINTER}.make_cast (l_buf), create {CE_VOID_POINTER}.make_cast (l_chunk_data), l_bytes_left.to_natural_32).do_nothing
				end
				l_i := l_i + 1
			end
			l_chunks.item.destroy.do_nothing
			Current.pos := pos + l_size
			Result := l_size
		end

	seek (a_offset: NATURAL_64): NATURAL_64
		local
			l_length: NATURAL_64
		do
			l_length := ((create {CE_CLASS_POINTER [PD_GRIDFILE]}.make_shared (c_pointer))).item.get_contentlength
			if l_length < a_offset then
				Current.pos := l_length
			else
				Current.pos := a_offset
			end
			Result := pos
		end

feature {PD_GRIDFILE} -- private routines

	bson_iterator_bin_len (a_i: CE_POINTER [PD_BSON_ITERATOR]): INTEGER_32
		require
			not a_i.is_void
		local
			l_tmp_2: INTEGER_32
		do
			if (bson_iterator_bin_type (a_i)).to_integer_32 = 2 then
				l_tmp_2 := (a_i.item.int_raw) - 4
			else
				l_tmp_2 := (a_i.item.int_raw)
			end
			Result := l_tmp_2
		end

	bson_iterator_bin_data (a_i: CE_POINTER [PD_BSON_ITERATOR]): CE_POINTER [INTEGER_8]
		require
			not a_i.is_void
		local
			l_tmp_2: CE_POINTER [INTEGER_8]
		do
			if (bson_iterator_bin_type (a_i)).to_integer_32 = 2 then
				l_tmp_2 := (a_i.item.value) + 9
			else
				l_tmp_2 := (a_i.item.value) + 5
			end
			Result := l_tmp_2
		end

	bson_iterator_bin_type (a_i: CE_POINTER [PD_BSON_ITERATOR]): INTEGER_8
		require
			not a_i.is_void
		do
			Result := ((a_i.item.value)[4])
		end

	bson_iterator_subobject (a_i: CE_POINTER [PD_BSON_ITERATOR]; a_sub: CE_POINTER [PD_BSON])
		require
			not a_i.is_void
			not a_sub.is_void
		do
			a_sub.item.init_data ((a_i.item.value)).do_nothing
			a_sub.item.reset
			a_sub.item.finished := 1
		end

feature {ANY}

	structure_size: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return sizeof(struct __anonstruct_gridfile_22);"
		end

	structure_alignment: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return __alignof__(struct __anonstruct_gridfile_22);"
		end

	offset_of (a_field_name: STRING_8): INTEGER_32
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_field_name)
			Result := c__offset_of (l_c_str.item)
		end

feature {ANY}

	copy (a_other: like Current)
		do
			make
			internal_c_pointer.memory_copy (a_other.c_pointer, structure_size)
		end

	is_equal (a_other: like Current): BOOLEAN
		do
			Result := internal_c_pointer.memory_compare (a_other.c_pointer, structure_size)
		end

feature {NONE}

	structure_info: CE_STRUCTURE
		do
			-- dummy for CE_CLASS invariant
			create Result.make_struct (<<>>)
			Result.alignment := structure_alignment
		end

	c__gfs (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfile_22 *)$a_c_pointer)->gfs;"
		end

	c__meta (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfile_22 *)$a_c_pointer)->meta;"
		end

	c__pos (a_c_pointer: POINTER): NATURAL_64
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_gridfile_22 *)$a_c_pointer)->pos;"
		end

	c__id (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)&((struct __anonstruct_gridfile_22 *)$a_c_pointer)->id;"
		end

	c__remote_name (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfile_22 *)$a_c_pointer)->remote_name;"
		end

	c__content_type (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfile_22 *)$a_c_pointer)->content_type;"
		end

	c__length (a_c_pointer: POINTER): NATURAL_64
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_gridfile_22 *)$a_c_pointer)->length;"
		end

	c__chunk_num (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_gridfile_22 *)$a_c_pointer)->chunk_num;"
		end

	c__pending_data (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfile_22 *)$a_c_pointer)->pending_data;"
		end

	c__pending_len (a_c_pointer: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return ((struct __anonstruct_gridfile_22 *)$a_c_pointer)->pending_len;"
		end

	c__set_gfs (a_c_pointer: POINTER; a_gfs: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->gfs = $a_gfs;"
		end

	c__set_meta (a_c_pointer: POINTER; a_meta: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->meta = $a_meta;"
		end

	c__set_pos (a_c_pointer: POINTER; a_pos: NATURAL_64)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->pos = $a_pos;"
		end

	c__set_id (a_c_pointer: POINTER; a_id: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->id = *(union __anonunion_bson_oid_t_12 *)$a_id;"
		end

	c__set_remote_name (a_c_pointer: POINTER; a_remote_name: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->remote_name = $a_remote_name;"
		end

	c__set_content_type (a_c_pointer: POINTER; a_content_type: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->content_type = $a_content_type;"
		end

	c__set_length (a_c_pointer: POINTER; a_length: NATURAL_64)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->length = $a_length;"
		end

	c__set_chunk_num (a_c_pointer: POINTER; a_chunk_num: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->chunk_num = $a_chunk_num;"
		end

	c__set_pending_data (a_c_pointer: POINTER; a_pending_data: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->pending_data = $a_pending_data;"
		end

	c__set_pending_len (a_c_pointer: POINTER; a_pending_len: INTEGER_32)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfile_22 *)$a_c_pointer)->pending_len = $a_pending_len;"
		end

	c__offset_of (a_field_name: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%", <string.h>, <stddef.h>, <stdio.h>"
		alias
			"[
				if(strcmp($a_field_name, "gfs") == 0) return offsetof(struct __anonstruct_gridfile_22, gfs);
				if(strcmp($a_field_name, "meta") == 0) return offsetof(struct __anonstruct_gridfile_22, meta);
				if(strcmp($a_field_name, "pos") == 0) return offsetof(struct __anonstruct_gridfile_22, pos);
				if(strcmp($a_field_name, "id") == 0) return offsetof(struct __anonstruct_gridfile_22, id);
				if(strcmp($a_field_name, "remote_name") == 0) return offsetof(struct __anonstruct_gridfile_22, remote_name);
				if(strcmp($a_field_name, "content_type") == 0) return offsetof(struct __anonstruct_gridfile_22, content_type);
				if(strcmp($a_field_name, "length") == 0) return offsetof(struct __anonstruct_gridfile_22, length);
				if(strcmp($a_field_name, "chunk_num") == 0) return offsetof(struct __anonstruct_gridfile_22, chunk_num);
				if(strcmp($a_field_name, "pending_data") == 0) return offsetof(struct __anonstruct_gridfile_22, pending_data);
				if(strcmp($a_field_name, "pending_len") == 0) return offsetof(struct __anonstruct_gridfile_22, pending_len);
				fprintf(stderr, "struct __anonstruct_gridfile_22.%s not found\n", (char *)$a_field_name);
				return -1;
			]"
		end

end
