class
	PD_GRIDFS

inherit
	CE_UTILITIES
		redefine
			copy,
			is_equal
		end
	CE_CLASS
		redefine
			make,
			make_shared,
			structure_size,
			offset_of,
			copy,
			is_equal
		end
	S_TIME_STATIC
		redefine
			copy,
			is_equal
		end
	S_STRING_STATIC
		redefine
			copy,
			is_equal
		end

create
	make,
	make_shared,
	make_initialized

feature {NONE}

	make
		do
			internal_c_pointer := internal_c_pointer.memory_alloc (structure_size)
			is_allocated_memory := True
		end

	make_shared (a_c_pointer: POINTER)
		do
			internal_c_pointer := a_c_pointer
		end

	make_initialized (a_data: TUPLE [
				a_client: CE_POINTER [PD_MONGO]
				a_dbname: CE_POINTER [INTEGER_8]
				a_prefix2: CE_POINTER [INTEGER_8]
				a_files_ns: CE_POINTER [INTEGER_8]
				a_chunks_ns: CE_POINTER [INTEGER_8]
			])
		require else
			a_data /= Void
		do
			make
			Current.client := a_data.a_client
			Current.dbname := a_data.a_dbname
			Current.prefix2 := a_data.a_prefix2
			Current.files_ns := a_data.a_files_ns
			Current.chunks_ns := a_data.a_chunks_ns
		end

feature {ANY} -- fields

	client: CE_POINTER [PD_MONGO] assign set__client
		do
			create {CE_CLASS_POINTER [PD_MONGO]} Result.make_shared (c__client (internal_c_pointer))
		end

	dbname: CE_POINTER [INTEGER_8] assign set__dbname
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__dbname (internal_c_pointer))
		end

	prefix2: CE_POINTER [INTEGER_8] assign set__prefix2
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__prefix2 (internal_c_pointer))
		end

	files_ns: CE_POINTER [INTEGER_8] assign set__files_ns
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__files_ns (internal_c_pointer))
		end

	chunks_ns: CE_POINTER [INTEGER_8] assign set__chunks_ns
		do
			create {CE_INTEGER_8_POINTER} Result.make_shared (c__chunks_ns (internal_c_pointer))
		end

feature {ANY} -- field setters

	set__client (a_client: CE_POINTER [PD_MONGO])
		do
			c__set_client (internal_c_pointer, a_client.c_pointer)
		end

	set__dbname (a_dbname: CE_POINTER [INTEGER_8])
		do
			c__set_dbname (internal_c_pointer, a_dbname.c_pointer)
		end

	set__prefix2 (a_prefix2: CE_POINTER [INTEGER_8])
		do
			c__set_prefix2 (internal_c_pointer, a_prefix2.c_pointer)
		end

	set__files_ns (a_files_ns: CE_POINTER [INTEGER_8])
		do
			c__set_files_ns (internal_c_pointer, a_files_ns.c_pointer)
		end

	set__chunks_ns (a_chunks_ns: CE_POINTER [INTEGER_8])
		do
			c__set_chunks_ns (internal_c_pointer, a_chunks_ns.c_pointer)
		end

feature {ANY} -- public routines

	insert_file (a_name: CE_POINTER [INTEGER_8]; a_id: PD_BSON_OID_T; a_length: NATURAL_64; a_contenttype: CE_POINTER [INTEGER_8]): INTEGER_32
		local
			l_id: CE_ARRAY [PD_BSON_OID_T]
			l_command: CE_ARRAY [PD_BSON]
			l_ret: CE_ARRAY [PD_BSON]
			l_res: CE_ARRAY [PD_BSON]
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_result: INTEGER_32
			l_d: INTEGER_64
			l_return: BOOLEAN
		do
			create {CE_CLASS_ARRAY [PD_BSON_OID_T]} l_id.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_command.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_ret.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_res.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)

			l_id.item := a_id.twin
			l_command.item.init
			l_command.item.append_oid (ce_cstring ("filemd5"), l_id).do_nothing;
			l_command.item.append_string (ce_cstring ("root"), prefix2).do_nothing
			l_command.item.finish.do_nothing
			l_result := (create {P_SRC_MONGO_STATIC}).mongo_run_command (client, dbname, l_command, l_res);
			l_command.item.destroy
			if l_result /= 0 then
				Result := l_result
				l_return := True
			end
			if not l_return then
				l_ret.item.init
				l_ret.item.append_oid (ce_cstring ("_id"), l_id).do_nothing
				if not a_name.is_void then
					if (a_name[0]).to_integer_32 /= 0 then
						l_ret.item.append_string (ce_cstring ("filename"), a_name).do_nothing
					end
				end
				bson_append_long (l_ret, ce_cstring ("length"), a_length.to_integer_64).do_nothing
				l_ret.item.append_int (ce_cstring ("chunkSize"), 262_144).do_nothing
				l_d := {INTEGER_64} 1_000 * (time (create {CE_INTEGER_32_POINTER}.make_shared (default_pointer))).to_integer_64
				l_ret.item.append_date (ce_cstring ("uploadDate"), l_d).do_nothing
				l_it.item.bson_find (l_res, ce_cstring ("md5")).do_nothing;
				l_ret.item.append_string (ce_cstring ("md5"), (l_it.item.string)).do_nothing;
				l_res.item.destroy
				if not a_contenttype.is_void then
					if (a_contenttype[0]).to_integer_32 /= 0 then
						l_ret.item.append_string (ce_cstring ("contentType"), a_contenttype).do_nothing
					end
				end
				l_ret.item.finish.do_nothing
				l_result := client.item.insert (files_ns, l_ret, create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer));
				l_ret.item.destroy
				Result := l_result
			end
		end

	store_buffer (a_data_0: CE_POINTER [INTEGER_8]; a_length: NATURAL_64; a_remotename, a_contenttype: CE_POINTER [INTEGER_8]): INTEGER_32
		local
			l_end: CE_POINTER [INTEGER_8]
			l_data_ptr: CE_POINTER [INTEGER_8]
			l_id: CE_ARRAY [PD_BSON_OID_T]
			l_chunk_number: INTEGER_32
			l_chunk_len: INTEGER_32
			l_o_chunk: CE_POINTER [PD_BSON]
		do
			create {CE_CLASS_ARRAY [PD_BSON_OID_T]} l_id.make (<<1>>)

			l_end := a_data_0 + a_length.to_integer_32
			l_data_ptr := a_data_0
			l_chunk_number := 0
			l_id.item.bson_oid_gen
			from
			until
				not ((l_data_ptr.c_pointer.to_integer_32).to_natural_32 < (l_end.c_pointer.to_integer_32).to_natural_32)
			loop
				if {NATURAL_32} 262_144 < (l_end.distance (l_data_ptr)).to_natural_32 then
					l_chunk_len := 262_144
				else
					l_chunk_len := ((l_end.distance (l_data_ptr)).to_natural_32).to_integer_32
				end
				l_o_chunk := (create {P_SRC_GRIDFS_STATIC}).chunk_new ((l_id.item).twin, l_chunk_number, l_data_ptr, l_chunk_len);
				client.item.insert (chunks_ns, l_o_chunk, create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer)).do_nothing;
				(create {P_SRC_GRIDFS_STATIC}).chunk_free (l_o_chunk)
				l_chunk_number := l_chunk_number + 1
				l_data_ptr := l_data_ptr + l_chunk_len
			end
			Result := (((create {CE_CLASS_POINTER [PD_GRIDFS]}.make_shared (c_pointer))).item.insert_file (a_remotename, (l_id.item).twin, a_length, a_contenttype))
		end

	remove_filename (a_filename: CE_POINTER [INTEGER_8])
		require
			not a_filename.is_void
		local
			l_query: CE_ARRAY [PD_BSON]
			l_files: CE_POINTER [PD_MONGO_CURSOR]
			l_file: CE_ARRAY [PD_BSON]
			l_it: CE_ARRAY [PD_BSON_ITERATOR]
			l_id: CE_ARRAY [PD_BSON_OID_T]
			l_b: CE_ARRAY [PD_BSON]
			l_break: BOOLEAN
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_query.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_file.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_ITERATOR]} l_it.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON_OID_T]} l_id.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_b.make (<<1>>)

			l_query.item.init;
			l_query.item.append_string (ce_cstring ("filename"), a_filename).do_nothing
			l_query.item.finish.do_nothing
			l_files := (create {P_SRC_MONGO_STATIC}).mongo_find (client, files_ns, l_query, create {CE_CLASS_POINTER [PD_BSON]}.make_shared (default_pointer), 0, 0, 0);
			l_query.item.destroy
			from
			until
				(l_break)
			loop
				l_break := False
				if (l_files.item.next) /= 0 then
					l_break := True
				end
				if not (l_break) then
					l_file.item := (l_files.item.current2).twin
					l_it.item.bson_find (l_file, ce_cstring ("_id")).do_nothing
					l_id.item := ((l_it.item.oid).item).twin
					l_b.item.init
					l_b.item.append_oid (ce_cstring ("_id"), l_id).do_nothing
					l_b.item.finish.do_nothing
					client.item.remove (files_ns, l_b, create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer)).do_nothing;
					l_b.item.destroy
					l_b.item.init
					l_b.item.append_oid (ce_cstring ("files_id"), l_id).do_nothing
					l_b.item.finish.do_nothing
					client.item.remove (chunks_ns, l_b, create {CE_CLASS_POINTER [PD_MONGO_WRITE_CONCERN]}.make_shared (default_pointer)).do_nothing;
					l_b.item.destroy
				end
			end
			l_break := False
			l_files.item.destroy.do_nothing
		end

	find_query (a_query: CE_POINTER [PD_BSON]; a_gfile: CE_POINTER [PD_GRIDFILE]): INTEGER_32
		local
			l_upload_date: CE_ARRAY [PD_BSON]
			l_final_query: CE_ARRAY [PD_BSON]
			l_out: CE_ARRAY [PD_BSON]
			l_i: INTEGER_32
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_upload_date.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_final_query.make (<<1>>)
			create {CE_CLASS_ARRAY [PD_BSON]} l_out.make (<<1>>)

			l_upload_date.item.init
			l_upload_date.item.append_int (ce_cstring ("uploadDate"), -1).do_nothing
			l_upload_date.item.finish.do_nothing
			l_final_query.item.init
			l_final_query.item.append_bson (ce_cstring ("query"), a_query).do_nothing
			l_final_query.item.append_bson (ce_cstring ("orderby"), l_upload_date).do_nothing
			l_final_query.item.finish.do_nothing
			l_i := (((create {P_SRC_MONGO_STATIC}).mongo_find_one (client, files_ns, l_final_query, create {CE_CLASS_POINTER [PD_BSON]}.make_shared (default_pointer), l_out)) = 0).to_integer;
			l_upload_date.item.destroy
			l_final_query.item.destroy
			if not l_i.to_boolean then
				Result := (-1)
			else
				gridfile_init ((create {CE_CLASS_POINTER [PD_GRIDFS]}.make_shared (c_pointer)), l_out, a_gfile).do_nothing
				l_out.item.destroy
				Result := 0
			end
		end

	find_filename (a_filename: CE_POINTER [INTEGER_8]; a_gfile: CE_POINTER [PD_GRIDFILE]): INTEGER_32
		require
			not a_filename.is_void
		local
			l_query: CE_ARRAY [PD_BSON]
			l_i: INTEGER_32
		do
			create {CE_CLASS_ARRAY [PD_BSON]} l_query.make (<<1>>)

			l_query.item.init
			l_query.item.append_string (ce_cstring ("filename"), a_filename).do_nothing
			l_query.item.finish.do_nothing
			l_i := ((create {CE_CLASS_POINTER [PD_GRIDFS]}.make_shared (c_pointer))).item.find_query (l_query, a_gfile);
			l_query.item.destroy
			Result := l_i
		end

feature {PD_GRIDFS} -- private routines

	bson_append_long (a_b: CE_POINTER [PD_BSON]; a_name: CE_POINTER [INTEGER_8]; a_i: INTEGER_64): INTEGER_32
		require
			not a_b.is_void
			not a_name.is_void
		local
			l_i: CE_ARRAY [INTEGER_64]
			l_return: BOOLEAN
		do
			create {CE_INTEGER_64_ARRAY} l_i.make (<<1>>)

			l_i.item := a_i
			if (a_b.item.append_estart (18, a_name, 8)) = -1 then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				a_b.item.append64 (create {CE_VOID_POINTER}.make_cast (l_i))
				Result := 0
			end
		end

	gridfile_init (a_gfs: CE_POINTER [PD_GRIDFS]; a_meta: CE_POINTER [PD_BSON]; a_gfile: CE_POINTER [PD_GRIDFILE]): INTEGER_32
		require
			not a_gfile.is_void
		local
			l_return: BOOLEAN
		do
			a_gfile.item.gfs := a_gfs
			a_gfile.item.pos := (0).to_natural_64
			a_gfile.item.meta := create {CE_CLASS_POINTER [PD_BSON]}.make_cast (((create {P_SRC_BSON_STATIC}).bson_malloc (((create {PD_BSON}.make).structure_size.to_natural_32).to_integer_32)))
			if (a_gfile.item.meta).is_void then
				Result := (-1)
				l_return := True
			end
			if not l_return then
				(create {P_SRC_BSON_STATIC}).bson_copy (a_gfile.item.meta, a_meta).do_nothing
				Result := 0
			end
		end

feature {ANY}

	structure_size: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return sizeof(struct __anonstruct_gridfs_21);"
		end

	structure_alignment: INTEGER_32
		external
			"C inline use %"libmongo.h%""
		alias
			"return __alignof__(struct __anonstruct_gridfs_21);"
		end

	offset_of (a_field_name: STRING_8): INTEGER_32
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_field_name)
			Result := c__offset_of (l_c_str.item)
		end

feature {ANY}

	copy (a_other: like Current)
		do
			make
			internal_c_pointer.memory_copy (a_other.c_pointer, structure_size)
		end

	is_equal (a_other: like Current): BOOLEAN
		do
			Result := internal_c_pointer.memory_compare (a_other.c_pointer, structure_size)
		end

feature {NONE}

	structure_info: CE_STRUCTURE
		do
			-- dummy for CE_CLASS invariant
			create Result.make_struct (<<>>)
			Result.alignment := structure_alignment
		end

	c__client (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfs_21 *)$a_c_pointer)->client;"
		end

	c__dbname (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfs_21 *)$a_c_pointer)->dbname;"
		end

	c__prefix2 (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfs_21 *)$a_c_pointer)->prefix;"
		end

	c__files_ns (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfs_21 *)$a_c_pointer)->files_ns;"
		end

	c__chunks_ns (a_c_pointer: POINTER): POINTER
		external
			"C inline use %"libmongo.h%""
		alias
			"return (EIF_POINTER)((struct __anonstruct_gridfs_21 *)$a_c_pointer)->chunks_ns;"
		end

	c__set_client (a_c_pointer: POINTER; a_client: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfs_21 *)$a_c_pointer)->client = $a_client;"
		end

	c__set_dbname (a_c_pointer: POINTER; a_dbname: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfs_21 *)$a_c_pointer)->dbname = $a_dbname;"
		end

	c__set_prefix2 (a_c_pointer: POINTER; a_prefix2: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfs_21 *)$a_c_pointer)->prefix = $a_prefix2;"
		end

	c__set_files_ns (a_c_pointer: POINTER; a_files_ns: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfs_21 *)$a_c_pointer)->files_ns = $a_files_ns;"
		end

	c__set_chunks_ns (a_c_pointer: POINTER; a_chunks_ns: POINTER)
		external
			"C inline use %"libmongo.h%""
		alias
			"((struct __anonstruct_gridfs_21 *)$a_c_pointer)->chunks_ns = $a_chunks_ns;"
		end

	c__offset_of (a_field_name: POINTER): INTEGER_32
		external
			"C inline use %"libmongo.h%", <string.h>, <stddef.h>, <stdio.h>"
		alias
			"[
				if(strcmp($a_field_name, "client") == 0) return offsetof(struct __anonstruct_gridfs_21, client);
				if(strcmp($a_field_name, "dbname") == 0) return offsetof(struct __anonstruct_gridfs_21, dbname);
				if(strcmp($a_field_name, "prefix2") == 0) return offsetof(struct __anonstruct_gridfs_21, prefix);
				if(strcmp($a_field_name, "files_ns") == 0) return offsetof(struct __anonstruct_gridfs_21, files_ns);
				if(strcmp($a_field_name, "chunks_ns") == 0) return offsetof(struct __anonstruct_gridfs_21, chunks_ns);
				fprintf(stderr, "struct __anonstruct_gridfs_21.%s not found\n", (char *)$a_field_name);
				return -1;
			]"
		end

end
